/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */

package org.aidworld.aidbase.document;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;

/**
 * Represents a HTML Document
 */

public class HtmlDocument implements Document {

    private String documentString = null;
    private String[] lines = null;
    private Map attributes;
    
	private static final Logger LOG = Logger.getLogger(HtmlDocument.class);

	/**
	 * Constructor which takes a string array
	 * 
	 * @param lines			File content
	 * @param attributes	File attributes
	 */
    public HtmlDocument(String[] lines, Map attributes) {
        this.lines = lines;
        this.attributes = attributes;
    }

    /**
     * Constructor which takes a single string
     * 
     * @param documentContent	File content
     * @param attributes		File attributes
     */
    public HtmlDocument(String documentContent, Map attributes) {
        documentString = documentContent;
        this.attributes = attributes;
    }

    /**
     * Constructor which takes a file
     * 
     * @param file			The html file
     * @param attributes	The html page attributes
     * @throws IOException	If any file operation fails 
     */
    public HtmlDocument(File file, Map attributes) throws IOException {

        StringBuffer stringBuffer = new StringBuffer();

        BufferedReader fileReader = new BufferedReader(new FileReader(file));

        int charRead;
        charRead = fileReader.read();
        while(charRead != -1) {
            stringBuffer.append((char)charRead);
            charRead = fileReader.read();
        }

        documentString = stringBuffer.toString();
        this.attributes = attributes;
    }

    /**
     * Replaces pieces of a string with an other string
     * 
     * @param string	The input string
     * @param search	The serch string
     * @param replace	The replace string
     * @return	The sring with the replaced values in it
     */
    public static String replaceAllRecursively(
        String string, String search, String replace) {

        String tempString = string;

        Pattern p = Pattern.compile(search);
        Matcher m = p.matcher(tempString);

        while(m.find()) {
            tempString = m.replaceAll(replace);
            m = p.matcher(tempString);
        }
        return tempString;
    }

    /**
     * Replaces pieces of a string with an other string
     * 
     * @param string	The input string
     * @param search	The serch string
     * @param replace	The replace string
     * @return	The sring with the replaced values in it
     */
    public static String replaceFirstRecursively(
        String string, String search, String replace) {

        String tempString = string;

        Pattern p = Pattern.compile(search);
        Matcher m = p.matcher(tempString);

        while(m.find()) {
            tempString = m.replaceAll(replace);
            m = p.matcher(tempString);
        }
        return tempString;
    }


    /**
     * Returns the document as a string
     * 
     * @return String representation of the document
     */
    public String toString() {
        if(documentString == null && lines != null) {
            // construct string from array
            StringBuffer buffer = new StringBuffer();
            for(int x=0; x<lines.length; ++x) {
                buffer.append(lines[x]);
            }
            documentString = buffer.toString();
        }
        return documentString;
    }

    /**
     * Returns the document as a string aray
     * 
     * @return String array representation of the document
     */
    public String[] getLines() {
        if(lines == null) {
            // construct array from string
            ArrayList linesList = new ArrayList();
            StringTokenizer tokenizer = new StringTokenizer(documentString, "\n");
            while(tokenizer.hasMoreTokens()) {
                linesList.add(tokenizer.nextToken());
            }

            lines = (String[]) linesList.toArray(new String[] {});
        }
        return lines;
    }

    /**
     * Returns the attributes of the document
     * 
     * @return The attributes
     */
    public Map getAttributes() {
        return attributes;
    }
    
    /**
     * Extracts the contents of the html header tag if one can be found,
     * (excluding the start head and end head tags themselves)
     * otherwise returns null.
     * 
     * @return the contents of the html header tag, or null.
     */
    public String getHtmlHeaderContents(){
    	
    	String str = this.toString();
    	
		// Try and extract the header
		Pattern pHead = Pattern.compile("(?s)(?i).*?<head.*?>(.*?)</head>.*?");
		Matcher mHead = pHead.matcher(str);
		String htmlHeader;
		if (mHead.matches()){
			
			htmlHeader = mHead.group(1);
			if (LOG.isDebugEnabled()) LOG.debug("Head tag found - html header successfully extracted");
			
		}else{
			
			htmlHeader = null;
			if (LOG.isDebugEnabled()) LOG.debug("No Head tag found! Assuming content has no header.");
			
		}
		return htmlHeader;
    }

	/**
	 * Extracts the contents of the html body tag if one can be found,
	 * (excluding the start body and end body tags themselves).
	 * If no body tag can be found, returns the contents of the html tags
	 * (with the header removed). If no html tags can be found, this method
	 * returns the entire contents of the document.
	 * 
	 * @return the contents of the html body tag if possible, otherwise more.
	 */
	public String getHtmlBodyContents(){
		
		String str = this.toString();
		
		// Try and extract the body
		Pattern pBody = Pattern.compile("(?s)(?i).*?<body.*?>(.*?)</body>.*?");
		Matcher mBody = pBody.matcher(str);
		String htmlBody;
		if (mBody.matches()){
			
			htmlBody = mBody.group(1);
			if (LOG.isDebugEnabled()) LOG.debug("Body tag found - body successfully extracted");
			
		}else{

			// No body tag found - take everything between the html tags 
			Pattern pHtml = Pattern.compile("(?s)(?i).*?<html.*?>(.*?)</html>.*?");
			Matcher mHtml = pHtml.matcher(str);
			if (mHtml.matches()){
				
				// Found the html tag. Only take its contents.
				htmlBody = mHtml.group(1);
				if (LOG.isDebugEnabled()) LOG.debug("No body tag found - using everything between html tags minus header");

				// Remove the header tag if the string contains one
				htmlBody = htmlBody.replaceFirst("(?s)(?i)<head.*?>(.*?)</head>","");

			}else{
				
				// Doesn't even appear to have html tags at all. 
				// Log a warning then take the entire string
				LOG.warn("Can't find html or body tags in page content - will use entire contents as body tag");
				htmlBody = str;
			}
		}
		
		return htmlBody;
	}

	String originalEncoding = null;
	public String getOriginalEncoding() { return originalEncoding; }
    public void setOriginalEncoding(String encoding) { 
    	originalEncoding = encoding; 
    }
    
    public byte[] getBytes() { return documentString.getBytes(); }
}






