/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */

package org.aidworld.aidbase.factory;


import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLDecoder;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.aidworld.aidbase.core.DocumentTypeChecker;
import org.aidworld.aidbase.document.BinaryDocument;
import org.aidworld.aidbase.document.Document;
import org.aidworld.aidbase.document.HtmlDocument;
import org.aidworld.aidbase.error.AidbaseException;
import org.aidworld.aidbase.error.InternalFaultException;
import org.aidworld.aidbase.error.UnknownTestPageException;
import org.aidworld.aidbase.error.UnsupportedProtocolException;
import org.aidworld.aidbase.request.Request;
import org.aidworld.aidbase.request.TestRequest;
import org.aidworld.aidbase.util.CompressCounter;
import org.apache.commons.httpclient.Header;
import org.apache.log4j.Logger;

/**
 * This factory creates Documents from URLs and Files according to some built-in
 * heuristics. For example, retrieving a document from Reliefweb, or retrieving
 * documents of content type application/x-pdf, might result in a document object
 * with a different class to the usual one (StringDocument) in future.
 *
 * TODO - should StandardDocumentFactory extend BinaryDocumentFactory rather
 * than the other way around?
 */
public class StandardDocumentFactory extends DocumentFactoryBase {
	private boolean binary = false;

    private static final Logger LOG = Logger.getLogger(StandardDocumentFactory.class);

    private DocumentTypeChecker checker = null;
    
    private static final Pattern metaContentTypeTag = 
    	Pattern.compile
        ("(?i)(?s)(<meta[^>]*http-equiv[^>]*content-type[^>]*>)");
    
    private static final Pattern metaContentTypeContent = 
    	Pattern.compile
        ("(?i)(?s)content\\s*=\\s*((?:['\"]([^'\"]*)['\"])|(?:.*))");

    private static final Pattern matchQuotedValue = 
    	Pattern.compile
        // ("(?i)(?s)\"([^\"]*)\"");
    	("(?i)(?s)['\"]([^'\"]*)['\"]");

    private static final Pattern metaContentTypeValueCharset =
    	Pattern.compile
        //("(?i)(?s)([^\"]*)(?:;\\s*charset\\s*=\\s*([^\"; ]+))?");
    	("(?i)(?s)(?:;\\s*charset\\s*=\\s*([^ ;]+))");

    public static String getEncodingFromContentType(String header) {
        Matcher matchContentTypeValue = metaContentTypeValueCharset.matcher(header);
        
        if (matchContentTypeValue.find()) {
            return matchContentTypeValue.group(1);
        }
        
        LOG.debug("Content-Type header does not specify character set: '"
                + header + "'");
        
        return null;
    }

    public StandardDocumentFactory(boolean binary){
    	this.binary = binary;
    }
    
    public StandardDocumentFactory(boolean binary,DocumentTypeChecker c){
    	this.binary  = binary;
    	this.checker = c;
    }
    
    public StandardDocumentFactory(){
    	this.binary = false;
    }
    
    public static String getContentTypeFromBody(String body) {
        Matcher matchContentTypeTag = 
        	metaContentTypeTag.matcher(body);
        
        if (! matchContentTypeTag.find()) {
            LOG.debug("no Meta Content-Type tag in document");
            return null;
        }        	
        
        String tag = matchContentTypeTag.group(1);
        
        Matcher matchContentTypeContent = 
        	metaContentTypeContent.matcher(tag);
        
        if (! matchContentTypeContent.find()) {
            LOG.warn("cannot parse Meta Content-Type tag: "+tag);
            return null;
        }        	
        
        String contentType = matchContentTypeContent.group(1);
        
        Matcher matchQuotes = matchQuotedValue.matcher(contentType);
        if (matchQuotes.matches()) {
        	contentType = matchQuotes.group(1); 
        }
        
        return contentType;
    }
    
    private String getEncodingFromData(String contentTypeHeader, String content) 
    {
    	String encoding = "";
        String contentType = contentTypeHeader;
        if (contentType != null) {
            String newEncoding = getEncodingFromContentType(contentType);
            if (newEncoding != null) {
                encoding = newEncoding;
                LOG.debug("encoding from Content-Type header: "+encoding+
                    " ("+contentType+")");
            } else {
                LOG.debug("no encoding in Content-Type header: "+contentType);
            }
        } else {
            LOG.debug("no content type header");
        }

        contentType = getContentTypeFromBody(content);
        if (contentType != null && encoding.length() == 0) {
        	String newEncoding = getEncodingFromContentType(contentType);
            if (newEncoding != null) {
                encoding = newEncoding;
                LOG.debug("encoding from Meta Content-Type tag: "+encoding);
            } else {
                LOG.debug("No encoding in content type '"
                		+contentType+"' from meta tag");
            }
        } else {
            LOG.debug("no Meta Content-Type tag in document");
        }
        
        if (encoding.length() == 0) {
        	encoding = "ISO-8859-1";
        	LOG.debug("No encoding found in document, assuming "+encoding);
        } else {
        	LOG.debug("Encoding found in document: " + encoding);
        }
        
        return encoding;
    }

	private boolean allowUnsafeUrls = false;
	public void setAllowUnsafeUrls(boolean allow) {
		this.allowUnsafeUrls = allow;
	}
	
	public String getFilePath(String path) {
		return "/org/aidworld/aidbase/testpages/" + path;
	}
	
	public URL getFileUrl(String path) {
		path = getFilePath(path);
		return getClass().getResource(path);
	}

    public Document service (Request request) throws AidbaseException {
        URL url;
        url = request.getURL();
        
        
        Map originalHeaders = request.getHeaders();

        if (originalHeaders == null) {
        	// @todo what should we do here? Can this ever happen?
            //return;
        }
        
        Vector originalCookies = new Vector();        
        for (Iterator oldHeaderKeys = originalHeaders.keySet().iterator(); 
            oldHeaderKeys.hasNext(); ) 
        {
            String name = (String)( oldHeaderKeys.next() );
            Object value = originalHeaders.get(name);
            if (name.equalsIgnoreCase("cookie")) {
                //m_log.debug(name+" is a Cookie header");
                originalCookies.add(value);
            }
        }
        
        /* Removes a surplus http:// */
        String currenturl = url.toString();
        /*
         * the following line has been commented out
         * it changes http://www.google.com to http:/%20/www.google.com
         * which is not what we want
         * Lucas 2005-07-21
         */
       // currenturl = WebUtils.convertURLToLobandFormat(currenturl);
        if(currenturl.startsWith("http://")){
        	currenturl = currenturl.substring(7);
        	
        	String decodedUrl;
        	try {
        		decodedUrl = URLDecoder.decode(currenturl,"UTF-8");
        		} 
        	catch (UnsupportedEncodingException e) {
        		throw new InternalFaultException("Unsupported encoding UTF-8", e);
        	}
        	
        	if (decodedUrl.startsWith("http://") ||
			    decodedUrl.startsWith("https://"))
        	{
        		try {
        			url = new URL(currenturl);
        		} catch (MalformedURLException e) {
        			throw new InternalFaultException(
        					"Created an invalid URL: currenturl", e);
        		}
        	}
        }

		/* TODO: safeness of URLs should be a property of the Request? */
        if (!(request instanceof TestRequest) &&
        	! url.getProtocol().equals("http") &&
        	! url.getProtocol().equals("https") &&
        	! url.getProtocol().equals("file") &&
			! url.getProtocol().equals("jar") &&
        	! this.allowUnsafeUrls) {
            throw new UnsupportedProtocolException(
            		"Only HTTP URLs are supported:" +
            		" ["+url.toString() +"]");
        }
        
        if (url.getProtocol().equals("file")) {
        	String path = url.getHost() + url.getPath();
        	
        	if (path.startsWith("test/")) {
        		try {
	        		url = new URL("file:/test/" + 
	        				path.replaceFirst(".*/",""));
        		} catch (MalformedURLException e) {
        			throw new InternalFaultException(
        					"Created an invalid URL: currenturl", e);
        		}	        		
        	} else {
        		path = path.replaceFirst("\\.\\.","");
        		path = path.replaceFirst("^/+", "");
	        	url = getFileUrl(path);
	        	if (url == null) {
	        		throw new UnknownTestPageException(
						"Test page not found: "+path,
						new FileNotFoundException(path));
	        	}
        	}
        }
        
        // Escape spaces if any are present
        try {
            if (url.toString().indexOf(" ") != -1) {
                url = new URL(url.toString().replaceAll(" ", "%20"));
                request.setURL(url);
            }
        } catch (MalformedURLException e) {
            throw new InternalFaultException(
            		"Exception when escaping spaces in url", e);
        }
        
        LOG.debug("retrieving headers starting");
        
        
        URLContents urlContents;
        if (url.getProtocol().equals("http") || 
        	url.getProtocol().equals("https"))
        {
        	urlContents = new HttpUrlContents(url, request);
        } else if (url.getProtocol().equals("file") &&
        		url.getPath().startsWith("/test/")) {
        	urlContents = new TestUrlContents(url, request);
        } else {
        	urlContents = new FileUrlContents(url, request);
        }

        LOG.debug("retrieving headers finished");

        // This will update a url like 'http://www.google.com/ads'
        // to 'http://www.google.com/ads/' which is important for
        // url - rewriting
    	request.setURL(urlContents.url);    
        
        
        if (urlContents.url.getProtocol().equals("file")) {
        	try {
        		String newUrl = "file:/" + 
					urlContents.url.getFile().replaceFirst(".*/", "");
        		request.setURL(new URL(newUrl));
        		LOG.warn("New URL changed to: "+newUrl);
        	} catch (MalformedURLException e) {
        		LOG.warn("Made an invalid file URL from "
        				+ urlContents.url.toString() + ": " + e);
        	}
        }
        
		if (checker != null)
		{
			checker.checkContents(urlContents);
		}
        
        Map newHeaders = new Hashtable();
        Header[] oldHeaders = urlContents.allHeaders;
        
        LOG.debug("Number of cookies in oldHeaders: " + urlContents.cookies.size());
        Vector newCookies = new Vector();
        
        for (int i = 0; i < urlContents.cookies.size(); i++)
        {   
        	// ONLY Set-Cookies that are new
        	if (!originalCookies.contains(urlContents.cookies.get(i)))
        	{
        		newCookies.add(urlContents.cookies.get(i));
        	}
        }
        
        newHeaders.put("Set-Cookie", newCookies);
        
        // LOG.debug("received "+oldHeaders.length+" headers");
        for (int i = 0; i < oldHeaders.length; i++) 
        {
        	// LOG.debug("received "+oldHeaders[i]+" header");
            String name = oldHeaders[i].getName();
            Object value = oldHeaders[i].getValue();
            if (name != null && value != null) {
            	if (!name.equalsIgnoreCase("set-cookie")){
            		newHeaders.put(name.toLowerCase(), value);
            	}
            	// m_log.debug(headerNames[i]+": "+value);
            }
        }

        LOG.debug("downloading page started");
        byte [] bytes = urlContents.getBytes();
        LOG.debug("downloading page finished");
        
        CompressCounter.writetoBuffer(1, bytes.length);
                
        if (binary) {
        	return new BinaryDocument(bytes,newHeaders);
        }
        
        String content = null;
        String defaultEncoding = "ISO-8859-1";
        
        try {
        	content = new String(bytes, defaultEncoding);      
        } catch (UnsupportedEncodingException e) {
            throw new InternalFaultException(
            		"Unsupported encoding "+defaultEncoding, e);
        }

        // detect character set, and convert to UTF-8, to agree with the
        // Meta Content-Type tag inserted by filter.xsl
        String encoding = getEncodingFromData(
        		urlContents.getContentTypeHeader(), content);
        // LOG.debug("got encoding from data: "+encoding);

        if (!(encoding.equalsIgnoreCase(defaultEncoding))) 
        {
        	LOG.info("Converting from "+defaultEncoding+" to "+encoding);
        
	        try {
	        	content = new String(bytes, encoding);
	        } catch (UnsupportedEncodingException e) {
	        	LOG.warn("Unsupported encoding: "+encoding);
	        }
        }
        
        HtmlDocument doc = new HtmlDocument(content, newHeaders);
        doc.setOriginalEncoding(encoding);
        return doc;
    }
}
