/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */
package org.aidworld.aidbase.filter;

import java.net.URL;
import java.util.HashMap;
import java.util.Map;


/**
 * This generates the loband version of a cookie.
 * 
 * The grammar for cookies:
 *
 * cookie    = NAME "=" VALUE *(";" cookie-av)
 * NAME      = attr
 * VALUE     = value
 * cookie-av = "Comment" "=" value
 *           | "Domain" "=" value
 *           | "Max-Age" "=" value
 *           | "Path" "=" value
 *           | "Secure"
 *           | "Version" "=" 1*DIGIT
 * attr      =       token
 * value     =       word
 * token     = 1*<any CHAR except CTLs or tspecials>
 * tspecials = "(" | ")" | "<" | ">" | "@"
 *           | "," | ";" | ":" | "\" | <">
 *           | "/" | "[" | "]" | "?" | "="
 *           | "{" | "}" | SP | HT
 */

public class AidCookie{
    protected String 
		name    = null,
        value   = null,
        domain  = null,
        path    = null;
    
    protected Map extraAttributes = new HashMap();
    
    protected boolean secure = false;
    protected URL origin;

    /**
     * Constructor which takes a URL
     * @param cookie	New cookie with the correct grammar
     * @param origin	URL where the cookie was set
     * @throws IllegalArgumentException	if cookie has incomplete grammar
     */
    public AidCookie (String cookie, URL origin)
    throws IllegalArgumentException
    {
        this.origin = origin;
        String [] av_pairs = cookie.split(";");

        for (int i = 0; i < av_pairs.length; i++) {
            String av_pair = av_pairs[i];
            String [] pair_parts = av_pair.split("=", 2);
            if (pair_parts.length < 1 || pair_parts.length > 2) {
                throw new IllegalArgumentException("cookie contains bad a-v pair: '"+av_pair+"'");
            }
            
            /* Remove any leading or trailing spaces around the cookie
             * Attribute-Value pairs.
             */
            for (int j = 0; j < pair_parts.length; j++) {
                pair_parts[j] = pair_parts[j].replaceFirst("^\\s+", "");
                pair_parts[j] = pair_parts[j].replaceFirst("\\s+$", "");
            }

            String attribute_name = pair_parts[0];
            String attribute_value = null;
            if (pair_parts.length == 2)
                attribute_value = pair_parts[1];

            if (i == 0) {
                if (attribute_value == null)
                    throw new IllegalArgumentException
                        ("cookie contains no value: '"+cookie+"'");
                name = attribute_name;
                value = attribute_value;
            } else if (attribute_name.equalsIgnoreCase("Domain")) {
                domain = attribute_value;
            } else if (attribute_name.equalsIgnoreCase("Path")) {
                path = attribute_value;
            } else {
            	extraAttributes.put(attribute_name, attribute_value);
            }
        }
    }
    /**
     * Constructor which takes domain and path
     * @param name	name of the cookie
     * @param value	value of the cookie
     * @param domain	domain of the host
     * @param path	path of the cookie
     */
    public AidCookie(String name, String value, String domain, String path){
    	this.name = name;
    	this.value = value;
    	this.domain = domain;
    	this.path = path;
    }
    /**
     * Returns the domain of the cookie
     * @return domain
     */
    public String getDomain() {
        if (domain != null) {
            return domain;
        }

        return origin.getHost();
    }
/**
 * Returns the path of the cookie
 * @return path
 */
    public String getPath() {
        if (path != null) {
            return path;
        }

        return origin.getPath().replaceAll("/[^/]+$", "/");
    }
    /**
     * Returns the name of the cookie
     * @return name
     */
    public String getName(){
    	return name;
    }
    /**
     * Returns the value of the cookie
     * @return value
     */
    public String getValue(){
    	return value;
    }
    
}