/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */

package org.aidworld.aidbase.filter;

import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import org.aidworld.aidbase.core.Timer;
import org.aidworld.aidbase.document.Document;
import org.aidworld.aidbase.error.AidbaseException;
import org.aidworld.aidbase.error.DocumentException;
import org.aidworld.aidbase.error.InternalFaultException;
import org.aidworld.aidbase.error.RequestException;
import org.aidworld.aidbase.request.Request;
import org.aidworld.aidbase.webui.util.WebUtils;
import org.apache.commons.httpclient.cookie.CookieSpecBase;
import org.apache.commons.httpclient.cookie.MalformedCookieException;
import org.apache.log4j.Logger;

/**
 * Changes the name, domain and path of cookies, so that they can be sent
 * back to the original site unmodified, when returned to Aidbase at a
 * later date.
 *
 */
public class CookieFilter extends BasicFilter {

    private final Logger m_log = Logger.getLogger(CookieFilter.class);

    /**
     * Decode any encoded cookies sent to us by the client,
     * find the ones relevant to this website, and add them to the
     * request to be passed on to the server.
     * @param request	The users request
     * @throws RequestException If the encoding is not supported
     */
    public void doDecodeRequestCookies (Request request) 
    throws AidbaseException
    {
        Map oldHeaders = request.getHeaders();

        if (oldHeaders == null) {
            m_log.debug("no headers in request");
            return;
        }
        
        Vector oldCookieHeaders = new Vector();
        Vector oldHeaderNames = new Vector();

        for (Iterator oldHeaderKeys = oldHeaders.keySet().iterator(); 
            oldHeaderKeys.hasNext(); ) 
        {
            String name = (String)( oldHeaderKeys.next() );
            Object value = oldHeaders.get(name);
            m_log.debug("request header "+name+": "+value);
            if (name.equalsIgnoreCase("cookie")) {
                //m_log.debug(name+" is a Cookie header");
                oldCookieHeaders.add(value);
                oldHeaderNames.add(name);
            }
        }

        if (oldCookieHeaders.size() == 0) {
            m_log.info("no cookie headers sent in request");
            return;
        }

        // make sure that no destination server gets our encoded cookies, 
        // which may be intended for other sites anyway
        for (int i = 0; i < oldHeaderNames.size(); i++) {
            String name = (String)( oldHeaderNames.get(i) );
            oldHeaders.remove(name);
        }
        
        // TODO handle multiple cookies sent in the same header.
        // multiple cookies are comma-separated in theory, but this
        // doesn't seem to work on Konqueror at least, and Google sets
        // a cookie with an Expires field containing a comma.
        Vector oldCookies = new Vector();

               
        for (int header = 0; header < oldCookieHeaders.size(); header++) {
            String thisCookieHeader = (String)( oldCookieHeaders.get(header) );
            String [] theseCookies = thisCookieHeader.split("[;]\\s*");
            for (int cookie = 0; cookie < theseCookies.length; cookie++) {
                String thisCookie = theseCookies[cookie];
                if (thisCookie == null) {
                    m_log.error("null cookie in array split from "+thisCookieHeader);
                } else {
                    oldCookies.add(theseCookies[cookie]);
                }
            }
        } 
        
        if (oldCookies.size() == 0) {
            m_log.warn("no cookies found in cookie headers");
            return;
        }
        
        Vector newCookies = new Vector();

        for (int n_cookie = 0; n_cookie < oldCookies.size(); n_cookie++) {
            String oldCookie = (String)( oldCookies.get(n_cookie) );
            AidCookie newCookie = null;
            
            try {
                newCookie = untranslate(oldCookie, request);
            } catch (UnsupportedEncodingException e) {
                throw new InternalFaultException(
                		"unsupported encoding UTF-8: "+e, e);
            }
            
            if (newCookie != null) {
                newCookies.add(newCookie);
            }
        }

        if (newCookies.size() == 0) {
            m_log.info("no encoded cookies to send");
            return;
        }
        
        /*String newCookieHeader = (String)(newCookies.get(0));

        for (int n_cookie = 1; n_cookie < newCookies.size(); 
            n_cookie++) 
        {
            newCookieHeader += "; " + (String)(newCookies.get(n_cookie));
        }
        */
        oldHeaders.put("AidBase Cookies", newCookies);
    }

    /**
     * Encodes any cookies sent by the server into Aidbase cookie format,
     * which allows them to be distinguished by site and path, causes
     * the client to properly replace old cookies with new ones, and protects
     * them from being damaged by the translation process.
     * @param req
     * @param doc
     * @throws RequestException
     * @throws DocumentException
     */
    
    void doEncodeResponseCookies(Request req, Document doc) 
    throws DocumentException {
        Map headers = doc.getAttributes();
        if (headers == null) {
            m_log.info("no headers in response");
            return;
        }

        String oldCookieHeaderName = null;
        
        Object oldCookieHeaderList = null;

    	for (Iterator oldHeaderKeys = headers.keySet().iterator();
            oldHeaderKeys.hasNext(); )
        {
            String name = (String)( oldHeaderKeys.next() );
            Object value = headers.get(name);
            m_log.debug("response header "+name+": "+value);
            if (name.equalsIgnoreCase("set-cookie")) {
            	if(oldCookieHeaderName != null){
            		m_log.warn("Multiple set cookie headers");
            		break;
            	}
                oldCookieHeaderName = name;
                oldCookieHeaderList = value;
                oldHeaderKeys.remove();
            }
        }
        
        if (oldCookieHeaderList == null) {
            m_log.info("no Set-Cookie header in response");
            return;
        }

		// Please note: Set-Cookie header is a vector,
		// all others are just Strings.
        Vector oldCookies = new Vector();
        Vector oldCookieHeaders = (Vector)(oldCookieHeaderList);
        CookieSpecBase spec = new CookieSpecBase();
        URL documentUrl = req.getURL();
        
        for (int i = 0; i < oldCookieHeaders.size(); i++) {
        	String header = (String)( oldCookieHeaders.get(i) );

            try {
            	int port = documentUrl.getDefaultPort();
            	if (documentUrl.getPort() != -1){
            		port = documentUrl.getPort();
            	}

//            	m_log.debug("DocumentUrl path = " + documentUrl.getPath());
//            	m_log.debug("DocumentUrl file = " + documentUrl.getFile());


            	org.apache.commons.httpclient.Cookie[] cookieArray 
            		= spec.parse(documentUrl.getHost(), port, 
            				documentUrl.getPath(), false, header);
            	
	            for (int c = 0; c < cookieArray.length; c++) {
	            	org.apache.commons.httpclient.Cookie oldCookie = cookieArray[c];
	            	String oldCookieString = oldCookie.toString();
	            	
	            	m_log.debug("Name for cookie is: " + oldCookie.getName() 
	            			    +" Domain is: " + oldCookie.getDomain());
	            	// if (oldCookie.getDomain().startsWith(".")) {
	            		oldCookieString += ";domain=" + oldCookie.getDomain();
	            	// }
	            	
	            	oldCookieString += ";path=" + oldCookie.getPath();
	            	
	            	
	            	if (oldCookie.getExpiryDate() != null){
	            		oldCookieString += ";expires=" + oldCookie.getExpiryDate();
	            	}
	            	
	            	oldCookies.add(oldCookieString);
	            }
            }
            catch (MalformedCookieException e) {
            	m_log.warn("Invalid cookie received: " + header);
            }
        }

        if (oldCookies.size() == 0) {
            m_log.warn("no valid cookies found in cookie header");
            return;
        }

        Vector newCookies = new Vector();

        for (int i = 0; i < oldCookies.size(); i++) {
            String oldCookie = (String)(oldCookies.get(i));
            m_log.debug("original cookie: " + oldCookie);
            try {
                String newCookie = translate(oldCookie, req);
                m_log.debug("translated cookie: " + newCookie);
                if (newCookie != null) {
                    newCookies.add(newCookie);
                }
            } catch (UnsupportedEncodingException e) {
                throw new DocumentException("unsupported encoding: "+e);
            }
        }

        if (newCookies.size() < 1) {
        	m_log.warn("Found no valid cookies, giving up");
        	return;
        }

        headers.put(oldCookieHeaderName, newCookies);
    }

    /**
     * @see org.aidworld.aidbase.core.Filter#transform(org.aidworld.aidbase.core.Document, org.aidworld.aidbase.core.Request)
     * For the specification for cookies, see @link http://www.cis.ohio-state.edu/cgi-bin/rfc/rfc2109.html
     *
     * The grammar for Set-Cookie headers:
     *
     * set-cookie = "Set-Cookie:" cookies
     * cookies    = 1#cookie
     * cookie     = NAME "=" VALUE *(";" cookie-av)
     * NAME       = attr
     * VALUE      = value
     * cookie-av  = "Comment" "=" value
     *            | "Domain" "=" value
     *            | "Max-Age" "=" value
     *            | "Path" "=" value
     *            | "Secure"
     *            | "Version" "=" 1*DIGIT
     */
    public Document service (Request request) throws AidbaseException {
        m_log.debug("starting CookieFilter.filter");

        Timer timer = new Timer();

        timer.push("Decoding request cookies");
        doDecodeRequestCookies(request);

        timer.pop().push("Servicing request");
        Document doc = super.service(request); 

        timer.pop().push("Encoding response cookies");
        doEncodeResponseCookies(request, doc);

        timer.pop();
        timer.logTimes(m_log);
        
        return doc;
    }

    /**
     * The cookie translation works as follows:
     * 
     * First we check that the site is allowed to set cookies for the
     * given domain. If present, the domain must start with a period ("."),
     * contain at least one more period, and be a suffix of the hostname
     * of the website, where the prefix (the remainder of the hostname)
     * does not contain any periods.
     * 
     * The cookie path must be a prefix of the path of the URL of
     * the page which set the cookie.
     * 
     * The cookie's domain, path and name are URL encoded and concatenated,
     * separated by colons (::). Prefixed by an underscore, this forms the
     * new encoded cookie's name. If the same site sets another cookie for the
     * same path with the same name, it will replace this cookie in the
     * user's browser, since the encoded cookie will also have the same name.
     * 
     * The cookie value is also URL-encoded, but otherwise the same as set.
     * 
     * The cookie attributes Comment, Max-Age and Secure are passed as-is
     * to the user's browser. They are never sent back to the originating
     * server anyway, and the client needs them to manage its cookies
     * effectively.
     * 
     * Any other attributes which the server sets on the cookie should
     * cause an exception. This may cause problems with newer versions
     * of the cookie standard.
     * 
     * @param cookieData
     * @param request
     * @return
     * @throws IllegalArgumentException
     * @throws UnsupportedEncodingException
     */
    
    private String translate(String cookieData, Request request)
    throws IllegalArgumentException, UnsupportedEncodingException
    {
        URL url = request.getURL();
        AidCookie oldCookie;
        try {
            oldCookie = new AidCookie(cookieData, url);
        } catch (Exception e) {
            m_log.error("Cookie error: "+e);
            return null;
        }

        // all encoded cookies start with underscores, 
        // cookies set by Aidbase for internal use do not.
        String newCookie = "_";
        
        // rewrite domain
        
        String oldDomain = oldCookie.getDomain();
        
        
        //preserve the initial . of domain cookies
        if(oldDomain.startsWith("."))
        {
        	
        	oldDomain = oldDomain.substring(1);
        	newCookie += ".";
        }
        
        String newDomain = WebUtils.convertURLToLobandFormat(oldDomain);
        
    	newCookie += URLEncoder.encode(newDomain, "UTF-8");
    	newCookie += "::";
    	newCookie += URLEncoder.encode(oldCookie.getPath(), "UTF-8");
    	newCookie += "::";
    	newCookie += oldCookie.name + "="
			+ oldCookie.value;
    	newCookie += ";path=" + WebUtils.getContextPath()
			+ WebUtils.getRealFilterServletPath() + "/" + newDomain
			+ oldCookie.getPath();

        for (Iterator extraAttributeNames 
        		= oldCookie.extraAttributes.keySet().iterator();
        	extraAttributeNames.hasNext(); ) {
        	
        	String name  = (String)(extraAttributeNames.next());
			String value = (String)(oldCookie.extraAttributes.get(name));
        	
        	if (value == null)
        		newCookie += "; " + name;
        	else if(!value.equalsIgnoreCase("path") && !value.equalsIgnoreCase("domain"))
        		newCookie += "; " + name + "=" + value;
        }

        return newCookie;
    }

    /**
     * Reverse the tranformations a cookie undertook when it was
     * translated in the method CookieFilter.translate(String cookieData, Request request).
     * Returns null if the cookie was set by 
     * Aidbase itself (recognisable by the fact they do not start with "_").
     * 
     * @param cookieData
     * @param request
     * @return
     * @throws IllegalArgumentException
     * @throws UnsupportedEncodingException
     */
    private AidCookie untranslate (String cookieData, Request request)
    throws IllegalArgumentException, UnsupportedEncodingException 
    {
    	m_log.debug("Cookie recu: " + cookieData);
        URL url = request.getURL();
        AidCookie oldCookie;
        
        try {
            oldCookie = new AidCookie(cookieData, url);
        } catch (Exception e) {
            m_log.error("received bad cookie from client: '"+cookieData+"'");
            return null;
        }

        if (! oldCookie.name.startsWith("_")) {
            // this is not an encoded cookie but one set by Aidbase itself.
            return null;
        }
        
        //preserve the dot from the beginning of domain cookies
        String sPrefix = "";
        String myData;
        if(oldCookie.name.startsWith("_."))
        {
        	sPrefix =".";
        	myData=oldCookie.name.substring(2);
        }
        else
        {
        	myData=oldCookie.name.substring(1);
        }
        m_log.debug("original request cookie: "+cookieData);

        String [] myDataValues = myData.split("::");
        if (myDataValues.length < 3) {
            m_log.warn("bad cookie: not enough fields "
                +"("+cookieData+")");
            return null;
        }
        
        String
            oldDomain = WebUtils.unConvertURLToLobandFormat(URLDecoder.decode(myDataValues[0], "UTF-8")),
            oldPath   = URLDecoder.decode(myDataValues[1], "UTF-8"),
			oldName   = URLDecoder.decode(myDataValues[2], "UTF-8"),
			oldValue  = URLDecoder.decode(oldCookie.value, "UTF-8");

  
        AidCookie newCookie = new AidCookie(oldName,oldValue,sPrefix + oldDomain,oldPath);
        m_log.debug("cookie is allowed");

        return newCookie;
    }

}
