/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */

package org.aidworld.aidbase.filter;

import java.net.URL;

import org.aidworld.aidbase.document.Document;
import org.aidworld.aidbase.document.HtmlDocument;
import org.aidworld.aidbase.error.AidbaseException;
import org.aidworld.aidbase.request.Request;

/**
 * This filter transforms an HTML string by converting all hyperlinks within the
 * HTML to ones that are redirected through the specified location. For instance
 * the HTML fragment: <p>
 *
 * <code>
 * &lt;a href="http://www.reliefweb.int/index"/&gt;
 * </code><p>
 *
 * would be converted into: <p>
 *
 * <code>
 *  &lt;a href=" http://www.aidworld.org/fetch?_ab_url=http://www.reliefweb.int/index" /&gt;
 * </code> <p>
 *
 * @todo Rething abstract method signatures! -NCW
 * @TODO url ENCODING
 */

public class LinkFilter extends BasicFilter {

	/* (non-Javadoc)
	 * @see org.aidworld.aidbase.core.Filter#transform(org.aidworld.aidbase.core.Document, org.aidworld.aidbase.core.Request)
	 */
	
	private static final String RELATIVE_URL = 
		"(([^h/'\"#][^'\">\\s]*)|([^/'\"][^':\">\\s]*))";
        
	private static final String FULL_URL =
		"(http://[^>'\"\\s]+)";

	private static final String ABSOLUTE_URL =
		"(/[^'\">\\s]*)";

    private static final String HYPER_LINK_START = 
        "(?i)(<a\\s[^>]*href\\s*=\\s*['\"]?)";

    private static final String HYPER_LINK_END = 
        "(['\"]?(>|\\s[^>]*>))";

    private static final String RELATIVE_URL_LINK =
        HYPER_LINK_START +  RELATIVE_URL + HYPER_LINK_END;

    private static final String FULL_URL_LINK =
        HYPER_LINK_START + FULL_URL + HYPER_LINK_END;

    private static final String ABSOLUTE_URL_LINK =
        HYPER_LINK_START + ABSOLUTE_URL + HYPER_LINK_END;

    private static final String AIDWORLD_URL_PREFIX = 
        "<a href='page?";

    private static final String AIDWORLD_URL_SECTION_SUFFIX =
        "request=text&url=";

    private static final String IMAGE_START = 
        "(?i)(<img\\s([^>]*)src\\s*=\\s*['\"]?)";

    private static final String IMAGE_END = 
        "(['\"\\s]?(>|(\\s[^>]*)>))";
    
    private static final String RELATIVE_URL_IMAGE =
        IMAGE_START + RELATIVE_URL + IMAGE_END;
    /* captures:
     *  <--- 2 --->   <---- 3 --->   <-- 3 --> */

    private static final String FULL_URL_IMAGE =
        IMAGE_START + FULL_URL + IMAGE_END;
    /*  <--- 2 --->   <- 1 -->   <-- 3 --> */

    private static final String ABSOLUTE_URL_IMAGE =
        IMAGE_START + ABSOLUTE_URL + IMAGE_END;
    /*  <--- 2 --->   <---- 1 --->   <-- 3 --> */

    private static final String AIDWORLD_IMAGE_TAG_START =
        "<img $2 src='ContentServlet?";

    private static final String AIDWORLD_IMAGE_TAG_MIDDLE =
        "param=http://";

    private static final String aidworldContentSection = AIDWORLD_IMAGE_TAG_START + 
        AIDWORLD_IMAGE_TAG_MIDDLE;

    private static final String BACKGROUND_START = 
        "(?i)(background\\s*=\\s*['\"])";

    private static final String BACKGROUND_END = 
        "(['\"\\s])";

    private static final String RELATIVE_BACKGROUND_ATTR =
        BACKGROUND_START + RELATIVE_URL + BACKGROUND_END;
    /* number of captures: 
     *  <-----  1 ----->   <---- 3 --->   <----- 1 ----> */

    private static final String FULL_BACKGROUND_ATTR =
        BACKGROUND_START + FULL_URL + BACKGROUND_END;
    /*  <----- 1 ------>   <-- 1 ->   <---- 1 -----> */

    private static final String ABSOLUTE_BACKGROUND_ATTR =
        BACKGROUND_START + ABSOLUTE_URL + BACKGROUND_END;
    /*  <----- 1 ------>   <---- 1 --->   <---- 1 -----> */

    private static final String aidworldBackgroundStart = "$1ContentServlet?" +
        "param=http://";

	public Document service(Request request)
    throws AidbaseException 
	{
        Document document = super.service(request);
        
        URL pageBaseURL = null;
        pageBaseURL = request.getURL();
        String hostName = pageBaseURL.getHost();
        
        // remove http:// from the head of the current page base URL,
        String fullPageSection = pageBaseURL.toString(); // .replaceAll("^http://", "");
        
        // and everything after the last /, to give us something we can prepend
        // to a relative link to make it absolute (all URLs sent to the Aidbase server
        // must be absolute).
        String currentDirectorySection = fullPageSection.replaceAll("/[^/]*$", "/");

        String source = document.toString();

        String filteredHTML = source.replaceAll(
            RELATIVE_URL_LINK, AIDWORLD_URL_PREFIX +  
            AIDWORLD_URL_SECTION_SUFFIX + currentDirectorySection + "/$2'>");

        filteredHTML = filteredHTML.replaceAll(
            FULL_URL_LINK, AIDWORLD_URL_PREFIX + 
            AIDWORLD_URL_SECTION_SUFFIX + "$2'>");

        filteredHTML = filteredHTML.replaceAll(
            ABSOLUTE_URL_LINK, AIDWORLD_URL_PREFIX +  
            AIDWORLD_URL_SECTION_SUFFIX + "http://" + hostName + "$2'>");

        // images (<img src="http://www.aidworld.org/blah.gif">, etc.)

        //@todo support preserving the border height and width attributes (all attributes) of the image tag.
        //eg <img border="3" src="indbul3a.gif" width="24" height="24">

        filteredHTML = filteredHTML.replaceAll(
            RELATIVE_URL_IMAGE, aidworldContentSection + currentDirectorySection + "/$3' $8>");

        filteredHTML = filteredHTML.replaceAll(
            FULL_URL_IMAGE, aidworldContentSection + "$3' $6>");

        filteredHTML = filteredHTML.replaceAll(
            ABSOLUTE_URL_IMAGE, aidworldContentSection + hostName + "$3' $6>");

        // background images

        // supports images set in the background attribute of the page
        // eg <body background="Mustique.jpg">
        // <table border="1" width="100%" background="Cancun.jpg">

        filteredHTML = filteredHTML.replaceAll(
            RELATIVE_BACKGROUND_ATTR, 
            aidworldBackgroundStart + currentDirectorySection + "/$2$5");

        filteredHTML = filteredHTML.replaceAll(
            FULL_BACKGROUND_ATTR, 
            aidworldBackgroundStart + "$2$3");

        filteredHTML = filteredHTML.replaceAll(
            ABSOLUTE_BACKGROUND_ATTR, 
            aidworldBackgroundStart + hostName + "$2$3");

        // don't accumulate slashes: two slashes not preceded
        // by a colon can always be replaced by one ( '//' -> '/' )
        // except in 'http://', which we leave alone by not matcing 
        // '//' preceded by a ':'

        /*  HYPER_LINK_START captures 1, HYPER_LINK_END 1 */
        filteredHTML = filteredHTML.replaceAll(
            HYPER_LINK_START + "([^'\"\\s]*)([^:])//+([^'\"\\s]*)" + 
            HYPER_LINK_END,
            "$1$2$3/$4$5");

        filteredHTML = filteredHTML.replaceAll(
            IMAGE_START + "([^'\"\\s]*)([^:])//+([^'\"\\s]*)" + 
            IMAGE_END,
            "$1$3$4/$5$6");
        
        filteredHTML = filteredHTML.replaceAll(
            BACKGROUND_START + "([^'\"\\s]*)([^:])//+([^'\"\\s]*)" + 
            BACKGROUND_END,
            "$1$2$3/$4$5");

        // for each ".." in the url, remove it and the directory 
        // specification preceding it.

        String dotDotURL = "([^'\"\\s]*)(/+[^.][^/]*/+\\.\\./+)([^'\"\\s]*)";

        filteredHTML = filteredHTML.replaceAll(
            HYPER_LINK_START + dotDotURL + HYPER_LINK_END,
            "$1$2/$4$5");

        filteredHTML = filteredHTML.replaceAll(
            IMAGE_START + dotDotURL + IMAGE_END,
            "$1$3/$5$6");
        
        filteredHTML = filteredHTML.replaceAll(
            BACKGROUND_START + dotDotURL + BACKGROUND_END,
            "$1$2/$4$5");

		return new HtmlDocument(filteredHTML, document.getAttributes());
	}
}
