/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */

package org.aidworld.aidbase.filter;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.Hashtable;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.aidworld.aidbase.document.Document;
import org.aidworld.aidbase.error.AidbaseException;
import org.aidworld.aidbase.error.DocumentException;
import org.aidworld.aidbase.request.Request;
import org.apache.log4j.Logger;

/** 
 * This filter finds and follows Meta Refresh tags in HTML documents.
 * @see http://www.htmlhelp.com/reference/html40/head/meta.html
 * @see http://www.htmlhelp.com/reference/wilbur/head/meta.html
 */

public class RedirectFilter extends BasicFilter {

	/* (non-Javadoc)
	 * @see org.aidworld.aidbase.core.Filter#transform(org.aidworld.aidbase.core.Document, org.aidworld.aidbase.core.Request)
	 */
    	
    private Logger log = Logger.getLogger(RedirectFilter.class);
    
	private static final String noScriptTagPattern = 
		"(?i)(?s)(<noscript>.*</noscript>)";
	private static final Pattern metaRefresh = 
	    Pattern.compile("(?i)(?s)(<meta[^>]*http-equiv[^>]*refresh[^>]*>)");
	private static final Pattern metaRefreshContent = 
	    Pattern.compile("(?i)(?s)content\\s*=\\s*['\"]([^'\"]*)['\"]");
	private static final Pattern metaRefreshUrl = 
	    Pattern.compile("(?i)(?s)(\\d+);.*url=([^'\"]*)");
	

	public Document service(Request request)
    throws AidbaseException
    {
        Document document = super.service(request);
        int depth = 0;
        
        String docString = document.toString();
        docString = docString.replaceAll(noScriptTagPattern, "");
        
        // if (log.isDebugEnabled())
        	log.debug("mangled = "+docString);
        
        Matcher metaRefreshMatcher = metaRefresh.matcher(docString);
        
        while (metaRefreshMatcher.find()) {
            depth++;
            if (depth > 10) {
                throw new DocumentException("too many meta refresh redirections");
            }
            
    		URL oldUrl;
 			oldUrl = request.getURL();
    
    		// ignore documents where Meta Refresh header is badly formatted
    		String tag = metaRefreshMatcher.group(1);
    		Matcher contentMatcher = metaRefreshContent.matcher(tag);
    		if (! contentMatcher.find()) {
                log.warn("meta refresh missing content tag: "+tag);
                return document;
    		}
    		
    		String content = contentMatcher.group(1);
    		Matcher urlMatcher = metaRefreshUrl.matcher(content);
    		if (! urlMatcher.find()) {
    		    log.info("meta refresh content missing URL: "+content);
    		    return document;
    		}
    
    		// don't follow long refreshes (> 99 seconds)
    		int refreshTime = Integer.parseInt(urlMatcher.group(1));
    		if (refreshTime > 99) {
                log.warn("meta refresh interval too long: "+refreshTime);
    			return document;
            }
    
    		// don't follow refreshes back to the same page
    		String newUrlString = urlMatcher.group(2);
    		if (oldUrl.toString().endsWith(newUrlString)) {
                log.warn("meta refresh to same URL: "+newUrlString);
    			return document;
            }
    
    		URL newUrl;
    
    		try {
    			newUrl = new URL(oldUrl, newUrlString);
                // log.debug("new URL based on: "+oldUrl+" and "+newUrlString);
                // log.debug("new URL is: "+newUrl.toString());
    		} catch (MalformedURLException e) {
    			// bad URL in refresh header
                log.warn("bad refresh URL: "+newUrlString);
    			return document;
    		}
    
            log.info("Redirecting to: "+newUrl.toString());
    		request.setURL(newUrl);
    		request.setMethod(Request.GET_METHOD);
    		request.setParameters(new Hashtable());
    		
    		try {
    			document = super.service(request);
    		} catch (Exception e) {
                throw new DocumentException(
                		"exception retrieving document from "+newUrl, e);
    		}
            
            if (document == null) {
                throw new DocumentException(
                		"failed to retrieve a document from "+newUrl);
            }

            /*
            log.debug("unmangled = "+document.toString());

            mangled = rfp.filter(request, 
            		new TestDocumentFactory(document.toString(), new Hashtable()));
            docString = mangled.toString();
            */

            docString = document.toString();

            if (docString == null) {
                throw new DocumentException("retrieved null document from "+newUrl);
            }
            
            docString = docString.replaceAll(noScriptTagPattern, "");
            
            // if (log.isDebugEnabled())
            // log.debug("mangled = "+docString);
            
            metaRefreshMatcher = metaRefresh.matcher(docString);
        }            
    		
        // log.debug("no more redirects");
            
    	return document;
	}
	
}


