/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */
package org.aidworld.aidbase.filter.rulefilter;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.aidworld.aidbase.filter.JavaScriptRewriter;
import org.aidworld.aidbase.webui.util.WebUtils;

import com.opensymphony.module.sitemesh.html.BasicRule;
import com.opensymphony.module.sitemesh.html.Tag;
import com.opensymphony.module.sitemesh.html.util.CharArray;

/**
 * This rule will filter out every attribute inside a tag that
 * we think is useless. See FilterList for all the attributes 
 * we keep. It will write everything to the main buffer.
 */
public class AttributeFilterRule extends BasicRule {
	Properties params = null;
	private boolean special_site = false;
	
	
	/**
	 * Main constructor
	 * 
	 * @param	tagNames 	All the HTML tags you want to 
	 * 						filter the attributes of
	 * @param	params		The general properties object which is 
	 * 						made in RuleFilter 
	 * @param	special		Name of the special site if we are 
	 * 						working on one, e.g. "hotmail".
	 */
	public AttributeFilterRule(String[] tagNames, Properties params, String special) {
		super(tagNames);
		this.params  = params;
		if(!special.equals("")){
			special_site = true;
		}
	}
	
	/**
	 * Retrieves all the attributes of a tag and passes them
	 * on to processAttribute.
	 * 
	 * @param	tag	 The tag that needs to be filtered
	 */
	public void process(Tag tag)
	{
		CharArray buffer = context.currentBuffer();
		
		if (tag.getType() == Tag.CLOSE) {
			tag.writeTo(buffer);
			return;
		}
		
		buffer.append("<");
		buffer.append(tag.getName());
		
		int attributeCount = tag.getAttributeCount();
		for (int i = 0; i < attributeCount; i++) {
			String name  = tag.getAttributeName(i);
			String value = tag.getAttributeValue(i);
			processAttribute(tag, name, value);
		}

		if (tag.getType() == Tag.EMPTY) 
			buffer.append("/");
		
		buffer.append(">");
	}
	
	private static String[] filteredAttributes = FilterList.ATTRIBUTES;
	private static String[] specialAttributes  = FilterList.SPECIAL_ATTRIBUTES;
	
	/**
	 * Removes the attribute from a tag if it isn't usefull 
	 * according to the list in FilterList.
	 * 
	 * @param tag	The tag which contains the attributes
	 * @param name	The name of the attribute
	 * @param value	The value of the attribute
	 */
	public void processAttribute(Tag tag, String name, String value)
	{
		boolean keepAttribute = false;
		if(params.get("aidbase.filter.disable")!=null)  // if filter is off
		{
			keepAttribute = true;
			if(name.equalsIgnoreCase("background"))
			{
				value = rewriteUrl(tag, value, (String)params.get("aidbase.param.type.binary"));
			}			
		}
		else // if filter is on
		{
			for (int i = 0; i < filteredAttributes.length; i++) {
				if(special_site){
					for (int j=0;j<specialAttributes.length;j++){
						if(specialAttributes[j].equalsIgnoreCase(name)){
							keepAttribute = true;
						}
					}
				}
				if (filteredAttributes[i].equalsIgnoreCase(name)){
					keepAttribute = true;
				}
					
			}
		}
		
		if(name.startsWith("on")){
			value = JavaScriptRewriter.rewrite(value,params);
		}
		
		
		String special = (String) (params.get("special"));
		
		if(keepAttribute){
			if(value != null){
				String lower = value.toLowerCase();
				if(special != null && !special.equals("yahoo_mail")){
					if(name.equals("value") && lower.startsWith("http")){
						try{
							value= WebUtils.getLobandCallUrl(value, "");
						}catch(Exception e){}
					}
				}
				// reliefweb uses the 'value' attribute of many of its 'input'tags to pass url's
				// this code makes sure that these url's are filtered properly through loband
				if(special != null && special.equals("reliefweb"))
				{
					if(tag.getName().equalsIgnoreCase("input") && name.equalsIgnoreCase("value")
							&& value.startsWith("/"))
					{
						try{
							value =  WebUtils.getLobandCallUrl("http://www.reliefweb.int" + value, "");
						
						} catch(Exception e) {}
					}
				}
			}
			CharArray buffer = context.currentBuffer();
			buffer.append(" ");
			buffer.append(name);
			if (value != null){
				buffer.append("=\"");
				buffer.append(value);
				buffer.append("\"");
			}
		}
	}
	
	
	/**
	 * This method rewrites the url to a loband url unless it is javascript.
	 * 
	 * @param tag				The input tag
	 * @param oldValue			The old url
	 * @param servletPath		The path of the application
	 * @param attribute Other attributes, specifying for
	 * 							 example which pipeline to use.
	 * @return The rewritten URL
	 */
	public String rewriteUrl(Tag tag, String oldValue, String parameters)
	{
		return rewriteUrl(tag, oldValue,
				WebUtils.getContextPath() + WebUtils.getRealFilterServletPath(),
				parameters);
	}
	
	/**  
	 * @param sURL eg. http://www.news.com/one/two/
	 * @return sURL with the last path element removed eg http://www.news.com/one/
	 */
	private String stripTrailingPath(String sURL)
	{
		int nPos;
		if(!sURL.endsWith("/"))
		{ // strip back to the path
			nPos = sURL.lastIndexOf('/', sURL.length());
			if(nPos>8)
			{
				sURL=sURL.substring(0, nPos+1);
			}			
		}
		if(sURL.endsWith("/"))
		{
			// there is possibly a path component to be stripped 
			nPos = sURL.lastIndexOf('/', sURL.length()-2); 
			if(nPos>7)
			{
				sURL = sURL.substring(0, nPos+1);
			}		
		}
		return sURL;
	}
	
	/**
	 * Takes URL of current page, and a link, and returns URL link is pointing to
	 * 
	 * This is different to the java new URL(sContext, sLink) because
	 * that doesnt handle links with too many ../ in them the same way
	 * that browsers do
	 * 
	 * @param sContext = URL of current page e.g. http://www.cnn.com/news.html
	 * @param sLink link we want to rewrite e.g. ../../hello.html or http://microsoft.com
	 * @return 
	 */
	public String newURLFromContext(String sContext, String sLink) throws MalformedURLException
	{
		String sNew;
		while(sLink.startsWith("../"))
		{
			sLink = sLink.substring(3);
			sContext = stripTrailingPath(sContext);
		}
		
		URL u = new URL(sContext);
		sNew = new URL(u, sLink).toString();
		return sNew;
	}
	
	/**
	 * This method rewrites the url to a loband url unless it is javascript.
	 * 
	 * @param tag				The input tag
	 * @param oldValue			The old url
	 * @param servletPath		The path of the application
	 * @param attribute Other attributes, specifying for
	 * 							 example which pipeline to use.
	 * @return The rewritten URL
	 */
	public String rewriteUrl(Tag tag, String oldValue, String servletPath, String parameters)
	{
		String absoluteUrl, newValue;
		
		
		/*
		 * change uppercase HTTP/Https to lowercase
		 */
		Pattern pat=Pattern.compile("http://",Pattern.CASE_INSENSITIVE);
		Matcher mat=pat.matcher(oldValue);
		if (mat.find()&&mat.start()==0)
			oldValue="http://"+oldValue.substring(7);
		
		Pattern pat2=Pattern.compile("https://",Pattern.CASE_INSENSITIVE);
		Matcher mat2=pat2.matcher(oldValue);
		if (mat2.find()&&mat2.start()==0)
			oldValue="https://"+oldValue.substring(8);
		
		
		//don't rewrite these link types,
		// just pass them straight through	
				
		if (oldValue.startsWith("https://"))
			return oldValue;
		if (oldValue.matches("(?s)\\s*[a-zA-Z]*:.*")  &&
				!(oldValue.startsWith("http:")) &&
				!(oldValue.startsWith("https:"))){
					if (oldValue.matches("(?s)\\s*javascript:.*")){
						return JavaScriptRewriter
							.rewriteLink(oldValue, params);
					}
					
					return oldValue;
		}
					
		if (oldValue.startsWith("#") ||
				oldValue.startsWith("http://www.aidbase.org"))
			return oldValue;

		oldValue = oldValue.replaceAll("&amp;","&");
		
		Pattern p = Pattern.compile("&#(\\d+);"); 
		Matcher m = p.matcher(oldValue);
		StringBuffer out = new StringBuffer();
		
		while (m.find()) 
		{
			String replacement;
			try {
				int code = Integer.parseInt(m.group(1));
				replacement = new String(new byte[]{(byte)code});
			} catch (NumberFormatException e) {
				replacement = "";
			}
			m.appendReplacement(out, replacement);
		}

		m.appendTail(out);
		oldValue = out.toString();
		
		String baseUrl = params.getProperty("base.url.full");
		
		// Get new URL from link and context, and check it is valid
		try 
		{
			absoluteUrl = newURLFromContext(baseUrl, oldValue);			
		} 
		catch (MalformedURLException e) 
		{
			//instead of throwing an exception, we just write the 
			//original URL of the link. this would normally make 
			//the user leave loband when they click on it
			//but since the URL is invalid, this won't happen
			
			return oldValue;
		}
		
		boolean isSecure = false;
					
		if (absoluteUrl.startsWith("https://") && 
				! baseUrl.startsWith("https://")) {
			newValue = 
				params.get("aidbase.server.root.https") +
				servletPath;
			isSecure = true;
		} else if (absoluteUrl.startsWith("http://") && 
				! baseUrl.startsWith("http://")) {
			newValue = 
				params.get("aidbase.server.root.http") +
				servletPath;
		} else {
			newValue = servletPath;
		}
		
		String endUrl = WebUtils.convertPortToParam(absoluteUrl);
		
		endUrl = WebUtils.convertURLToLobandFormat(endUrl);
		
		if(endUrl.startsWith("http://"))
		{
			endUrl = endUrl.substring(7);
		} else if (endUrl.startsWith("https://")){
			endUrl = endUrl.substring(8);
		}
		newValue += "/" + endUrl;

		boolean hasQuestionMark = (endUrl.indexOf('?') != -1);
		// set to true as soon as a question mark (showing beginning
		// of parameters list) is added to the rewritten url,
		// to prevent adding another one
		
		if(!parameters.equals(""))
		{
			if(hasQuestionMark)
			{
				newValue += "&";
			} else {
				newValue += "?";
				hasQuestionMark = true;
			}
			newValue += parameters;
		}
		if(isSecure)
		{
			if(hasQuestionMark)
			{
				newValue += "&";
			} else {
				newValue += "?";
				hasQuestionMark = true;
			}
			newValue += "_ab_secure=yes";
		}
			
		String x= params.get("aidbase.simulatorOnOff").toString();					
		if(x.equals("true"))
		{
			if(hasQuestionMark)
				newValue += "&";
			else
			{
				newValue += "?";
				hasQuestionMark = true;
			}
			newValue += params.get("aidbase.paramname.simulatorkey")
			+ "=" + params.get("aidbase.simulatorkey") + "&" 
			+ params.get("aidbase.paramname.simulatorbps") + "="
			+ params.get("aidbase.simulatorbps");
		}
		if(params.get("aidbase.filter.disable")!=null){
			if(hasQuestionMark)
				newValue += "&";
			else
			{
				newValue += "?";
				hasQuestionMark = true;
			}
			newValue += "_ab_nofilter=a";
		}
		
		String tagName = tag.getName();
		boolean showHeader = true;
		
		if ("0".equals(params.get("aidbase.header.enable"))) {
			showHeader = false;
		}
		
		if (tagName.equalsIgnoreCase("frame")
				|| tagName.equalsIgnoreCase("iframe")) {
			if (absoluteUrl.indexOf("_ab_locinfo=") != -1) {
				showHeader = true;
			} else {
				showHeader = false;
			}
		}
		else if (tag.hasAttribute("target")){
			if (tag.getAttributeValue("target").equals("_blank") 
					|| tag.getAttributeValue("target").equals("_top") 
					|| tag.getAttributeValue("target").equals("_new")){
				showHeader = true;
			}
		}
		
		if (!showHeader) {
			if(hasQuestionMark)
				newValue += "&";
			else
			{
				newValue += "?";
				hasQuestionMark = true;
			}

			newValue += params.get("aidbase.param.noheader")
				+ "=1";
		}
		
		return newValue;
	}
}
