/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */
package org.aidworld.aidbase.handler;

import java.net.URL;
import java.util.Iterator;
import java.util.Vector;

import org.aidworld.aidbase.core.Timer;
import org.aidworld.aidbase.document.Document;
import org.aidworld.aidbase.error.DocumentException;
import org.aidworld.aidbase.error.GoDirectException;
import org.aidworld.aidbase.filter.Stage;
import org.aidworld.aidbase.request.Request;
import org.aidworld.aidbase.util.Utils;
import org.apache.log4j.Logger;
import org.apache.log4j.NDC;

/**
 * TODO: Javadoc
 */
public class RetryHandler extends Thread implements RequestHandler {
	private Request   request   = null;
	private Stage     server    = null;
	private boolean   postRequest = false;
	private boolean   running, finished;
	private Throwable exception = null;
	private Document  document  = null;
	private Timer     timer     = new Timer();
	private String    name      = null;
	private Object    clientMonitor = null;
	
    private static Logger LOG = Logger.getLogger(RetryHandler.class);

	public RetryHandler(
			String name,
			Request request,
			boolean isPost, 
			Stage server,
			Object monitor)
	throws Exception 
	{
		super("RetryHandler");
		this.name = name;
		this.server = server;
		this.request = request.copy();
		this.postRequest = isPost;
		this.clientMonitor = monitor;
    	// request.setAttribute("request", arq);
	}
	
	public boolean   isRunning()    { return running; }
	public boolean   isFinished()   { return finished; }
	public Throwable getException() { return exception; }
	public Document  getDocument()  { return document; }
	
	private long startTimeMillis = 0;
	public long getStartTimeMillis() {
		return startTimeMillis;
	}
	
	public void run() {
		if (running) 
			throw new IllegalThreadStateException(
				"thread reused while still running");

		NDC.push(this.name);
		Utils.addHandler(this);

		exception = null;
		finished  = false;
		running   = true;
		Object myThreadMonitor = new Object();
		int seq = 0;
		
		startTimeMillis   = System.currentTimeMillis();
		int timeoutMillis = Utils.getMaxTimeoutMillis();
		
		long deadline   = startTimeMillis + timeoutMillis;
		long nextRetry  = startTimeMillis;
		long timeNow    = startTimeMillis;
		int  retries    = 5;
		int  retryDelay = 1;
		boolean started = false;
		
		Vector runningHandlers  = new Vector(), 
			finishedHandlers = new Vector();
		
		RequestHandler winner = null;

		while (System.currentTimeMillis() < deadline && winner == null) {
			timeNow = System.currentTimeMillis();

			long waitTime = (nextRetry - timeNow);
			if (waitTime > deadline - timeNow) 
				waitTime = deadline - timeNow;
			
			if (waitTime > 0) {
				timer.push("waiting "+waitTime+" ms for "
						+runningHandlers.size()+" threads");

				try {
					synchronized(myThreadMonitor) {
						myThreadMonitor.wait(waitTime);
					}
				} catch (InterruptedException e) {
					this.exception = e;
				}
				
				timer.pop();
			}
			
			if (isInterrupted()) {
				LOG.warn("thread interrupted, giving up");
				break;
			}

			timeNow = System.currentTimeMillis();

			for (Iterator i = runningHandlers.iterator(); i.hasNext(); ) {
				RequestHandler rh = (RequestHandler)( i.next() );
				if (!rh.isFinished()) 
					continue;
				
				finishedHandlers.add(rh);
				i.remove();
					
				Throwable rex = rh.getException();
				if (rex == null || rex instanceof GoDirectException) {
					// finished without exception, we have a winner!
					LOG.warn("Handler "+rh.getName()+
							" finished successfully, we have a winner!");
					winner = rh;
					break;
				}

				LOG.warn("Handler "+rh.getName()+" finished with an error: "+rex);
			}
			
			if (winner != null) break;
			
			int numRunning = runningHandlers.size();
			boolean startThread = false;
			
			if (numRunning == 0 && retries == 0) {
				LOG.warn("Document not complete and no retries remaining, " +
					"no threads running, giving up.");
				break;
			} else if (numRunning == 0) {
				if (!started) {
					LOG.debug("Starting initial download thread");
				} else {
					LOG.warn("Document not complete and no threads running," +
						"starting another now.");
				}
				startThread = true;
			} else if (retries == 0 && timeNow > deadline) {
				LOG.warn("Document not complete and out of time, giving up");
			} else if (retries == 0) {
				LOG.warn("Document not complete and no retries remaining, " +
					"just waiting now ("+(deadline-timeNow)+" ms)");
				nextRetry = deadline;
			} else if (timeNow >= nextRetry) {
				LOG.warn("Document not complete and no threads succeeded, "+
					"starting another now.");
				startThread = true;
			}
			
			if (!startThread) 
				continue;
			
			// try again (start another thread)
			retries--;
			started = true;
			
			try {
				RequestHandler rh = new StandardRequestHandler(name,
						request, postRequest, server, myThreadMonitor);
				runningHandlers.add(rh);
				rh.start();
				nextRetry = timeNow + retryDelay * 1000;
				retryDelay *= 4;
				LOG.warn("created a new thread ("+seq+"), now waiting for "
						+runningHandlers.size()+" threads");
				seq++;
			} catch (Exception e) {
				this.exception = e;
			}
		}
		
		if (winner != null) {
			this.document  = winner.getDocument();
			this.exception = (GoDirectException)( winner.getException() );
			this.request.setURL(winner.getRequest().getURL());
		} else {
			if (this.exception == null) 
			{
				for (Iterator i = finishedHandlers.iterator(); i.hasNext(); ) {
					RequestHandler rh = (RequestHandler)( i.next() );
					if (rh.getException() != null) {
						this.exception = rh.getException();
						break;
					}
				}
			}
			if (this.exception == null) 
			{
				this.exception = new DocumentException("Request timed out");
			}
		}

		synchronized(clientMonitor) {
			finished = true;
			running  = false;
			clientMonitor.notify();
		}

		/* stop all threads */
		for (Iterator i = runningHandlers.iterator(); i.hasNext(); ) {
			RequestHandler rh = (RequestHandler)( i.next() );
			if (rh.isRunning() && ! rh.isFinished()) {
				LOG.warn("stopping thread "+rh.toString());
				rh.interrupt();
			}
		}
		
		timer.logTimes(LOG);

		NDC.remove();
		Utils.removeHandler(this);
	}
	
	public String getURL() {
		if (request == null) return null;
		URL u = request.getURL();
		if (u == null) return null;
		return u.toString();
	}
	
	public String getCurrentTaskName() {
		return timer.getCurrentTaskName();
	}
	
	public Request getRequest() { return request; }
}
