/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */
package org.aidworld.aidbase.handler;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;

import javax.servlet.http.HttpServletRequest;

import org.aidworld.aidbase.document.Document;
import org.aidworld.aidbase.error.AidbaseException;
import org.aidworld.aidbase.error.DocumentException;
import org.aidworld.aidbase.error.DocumentNotFoundException;
import org.aidworld.aidbase.error.GoDirectException;
import org.aidworld.aidbase.error.MissingParameterException;
import org.aidworld.aidbase.error.UnsupportedProtocolException;
import org.aidworld.aidbase.filter.Stage;
import org.aidworld.aidbase.request.Request;
import org.aidworld.aidbase.util.Utils;
import org.aidworld.aidbase.webui.simulator.ThrottleFilter;
import org.aidworld.aidbase.webui.util.WebUtils;
import org.apache.log4j.Logger;
import org.apache.log4j.NDC;

/**
 * TODO: Javadoc
 */
public class StandardRequestHandler 
extends Thread 
implements RequestHandler 
{
	private Request   request;
	private Stage     server;
	private boolean   postRequest;
	private boolean   running, finished;
	private String    urlString = null;
	private Throwable exception = null;
	private Document  document  = null;
	private String    name      = null;
	private Object    clientMonitor = null;
	
    private static Logger LOG = Logger.getLogger(StandardRequestHandler.class);

    /**
     * Creates a standard request handler, which makes a single attempt to retrieve
     * and process a URL, giving up after the standard timeouts.
     * 
	 * @throws FileNotFoundException if the application properties file can't be found
	 * @throws IOException if the application properties file can't be read
	 * @throws IllegalArgumentException if a header is duplicated
	 */
    
	public StandardRequestHandler(
			String name,
			Request request,
			boolean isPost, 
			Stage server,
			Object monitor)
	throws IllegalArgumentException
	{
		super(name);

		this.request = request.copy();

		this.name    = name;
		this.server  = server;
		this.postRequest = isPost;
		this.clientMonitor = monitor;
    	// request.setAttribute("request", request);
	}
	
	public boolean   isRunning()    { return running; }
	public boolean   isFinished()   { return finished; }
	public Throwable getException() { return exception; }
	public Document  getDocument()  { return document; }
	
	private long startTimeMillis = 0;
	public long getStartTimeMillis() {
		return startTimeMillis;
	}
	
	public void run() {
		startTimeMillis = System.currentTimeMillis();
		Utils.addHandler(this);
    	NDC.push(name);
		
		try {
			if (running) 
				throw new IllegalThreadStateException(
					"thread reused while still running");
			
			exception  = null;
			finished = false;
			running  = true;
			
			handleRequest();

			// This attribute is read by tomcat to put it in the
			// log file tomcat/logs/???_log.date. This time is
			// the time it took in milliseconds to filter the request
			// through loband
			if(request.getRequest() != null)
				request.getRequest().setAttribute("LobandProcessingTime",
						""+(System.currentTimeMillis() - startTimeMillis));
		}
		catch (Throwable t)
		{
			request.getTimer().push("throwing exception: "+t);
			this.exception = t;
			
			LOG.warn("Error creating document: "+t);
			StackTraceElement[] trace = t.getStackTrace();
			for (int i = 0; i < trace.length; i++) {
				LOG.debug(i + ": " + trace[i].toString());
			}
		}

		request.getTimer().logTimes(LOG);
		
		synchronized(clientMonitor) {
			finished = true;
			running  = false;
			clientMonitor.notify();
		}
		
		Utils.removeHandler(this);
		NDC.remove();
	}
	
	/**
	 * Handles the request in the current thread, without monitoring 
	 * or exception handling.
	 */ 
	public void handleRequest() throws AidbaseException
	{
		String[] urlParams = null;
		String urlParamName = WebUtils.getAidbaseUrlParam();
		
		if (request.getParameter(urlParamName) instanceof String[])
			urlParams = (String[])(request.getParameter(urlParamName));
        
		if (urlParams == null || urlParams.length == 0)
            throw new MissingParameterException(
            		"You must supply the parameter "+
            		"'" + urlParamName + "'");
        
		String urlParam = urlParams[0];
		//Temporally solution for HttpUnit ssl testing
		urlParam = urlParam.replaceAll("^https:/([^/])", "https://$1");
		urlParam = urlParam.replaceAll("^http:/([^/])", "http://$1");
        
		if (urlParam == null || urlParam.equals(""))
            throw new MissingParameterException(
            		"You must supply the parameter "+
            		"'" + urlParamName + "'");

    	LOG.debug("Page request starting: " + urlParam);
    	urlString = urlParam;

        URL url;

        try 
		{
            url = new URL(urlParam);
        } 
        catch (MalformedURLException e) 
		{
            // The urlParam string isn't a valid url.
            // Try appending http to the front before rejecting it though,
            // as this is a frequest user error.

        	if (urlParam.indexOf("://") != -1)
        	{
        		throw new UnsupportedProtocolException(
        				"Only HTTP URLs are supported:" +
                		" ["+urlParam +"]");
        	}

        	try {
            	URL newurl = new URL("http://" + urlParam);
            	
            	request.getTimer().push("testing repaired URL");
                InputStream stream = newurl.openStream();
                stream.close();
                request.getTimer().pop();

                // If there was no exception above, this is a valid url.
                url = newurl;
            } catch (IOException newException) {
                // Appending the protocol didn't work - throw the original exception
                throw new DocumentNotFoundException(
                		"Unable to recognise URL: "+urlParam, e);	
            }
        }

        urlString = null;
        request.setURL(url);

        HttpServletRequest srq = (HttpServletRequest)( request.getRequest() );
        if (srq != null) { 
        	String accessedUrl = srq.getRequestURL().toString();
        
	        if (url.getProtocol().equals("https") && 
	       		accessedUrl.startsWith("http:"))
	       	{
	        	LOG.debug("Page request interrupted to go direct: " + urlParam);
	        	try
				{
	        		throw new GoDirectException(WebUtils.getLobandCallUrl(url.toString(), ""));
				} catch(IOException e) {}
	       	}
	        else if (url.getProtocol().equals("http") &&
	       		accessedUrl.startsWith("https:"))
	       	{
	        	LOG.debug("Page request interrupted to go direct: " + urlParam);
	        	try
				{
	        		throw new GoDirectException(WebUtils.getLobandCallUrl(url.toString(), ""));
				} catch(IOException e) {}
	       	}
        }
        
        if (postRequest) {
        	request.setMethod(Request.POST_METHOD);
        }

        // Check to see if request is simulated at low bandwidth
        // store parameters for later link re-writing
        int nSimulatorBps=0;
        int nSimulatorKey=0;
        
        if(request.getParameter(ThrottleFilter.BANDWIDTH_PARAM)!= null) {
        	String sSimulatorBps=((String[])(request.getParameter
        				(ThrottleFilter.BANDWIDTH_PARAM)))[0];        	
        	try {
                nSimulatorBps = Integer.parseInt(sSimulatorBps);
            } catch(NumberFormatException e) {
                if (LOG.isDebugEnabled()) LOG.debug("Bandwidth parameter invalid - "+e);                
            }
        }
        if(request.getParameter(ThrottleFilter.THROTTLER_KEY_URL_PARAM)!=null){
        	String sSimulatorKey=((String[])(request.getParameter
    				(ThrottleFilter.THROTTLER_KEY_URL_PARAM)))[0];
        	try {
                nSimulatorKey = Integer.parseInt(sSimulatorKey);
            } catch(NumberFormatException e) {
                if (LOG.isDebugEnabled()) LOG.debug("Bandwidth key parameter invalid - "+e);                
            }
        }
        if(nSimulatorBps >= 300)
        {
        	request.setSimulatorOnOff(true);
        	request.setSimulatorKey(nSimulatorKey);
        	request.setSimulatorBps(nSimulatorBps);
        }
        
        request.setFiltering(true);
        if(request.getParameter("_ab_nofilter")!=null)
        {
        	request.setFiltering(false);
        }
        
        // String requestType = request.getParameter("request");

        this.document = server.service(request);

        if (interrupted()) {
        	LOG.debug("Page request interrupted: " + urlParam);
        	return;
        }
        
        if (document == null) {
        	LOG.debug("Page request finished but returned no document: " + 
        			urlParam);
        	throw new DocumentException("Document Factory returned no document");
        }

    	LOG.debug("Page request succeeded: " + urlParam);
	}
	
	public String getURL() {
		if (urlString != null) return urlString;
		if (request   == null) return null;
		URL u = request.getURL();
		if (u == null) return null;
		return u.toString();
	}

	public String getCurrentTaskName() {
		return request.getTimer().getCurrentTaskName();
	}
	
	public Request getRequest() {
		return request;
	}
}
