/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */

package org.aidworld.aidbase.request;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import java.util.Properties;

import javax.servlet.http.HttpServletRequest;

import org.aidworld.aidbase.core.Timer;
import org.aidworld.aidbase.error.RecursionDepthException;
import org.aidworld.aidbase.error.RecursionNestingException;
import org.aidworld.aidbase.util.PartInfo;
import org.aidworld.aidbase.util.Utils;
import org.apache.commons.httpclient.HttpState;

/**
 * This abstract class represents the base from which Aidbase
 * requests are derived. The request is a protocol-independent way
 * to tell Aidbase which piece of information to retrieve from the
 * Internet (with a URL) and what kind of processing to perform.
 * 
 */
abstract class BasicRequest implements Request {
	private Map     headers;
    private int     depth = 0;
    private static final int maxDepth = 25;
    private String  contextPath     = null;
    private URL     url             = null;
    private Method  method          = Request.GET_METHOD;
    private Map     parameters      = null;
	private Map     filter_params   = new HashMap();
	private String  httpsHost       = null;
	private String  httpHost        = null;
	private int	    simulatorBps	= 0;
	private int	    simulatorKey	= 0;
	private boolean	simulatorOnOff  = false;
	private String  requestType     = "";
	private boolean bFiltering      = true;
    private Timer   timer           = new Timer();
	private HttpState  state;
    private PartInfo[] parts        = null;
    
	protected BasicRequest(boolean fudge) 
	{
		if (!fudge) { }
	}

	/**
	 * Create a BasicRequest
	 * 
	 * @throws FileNotFoundException if the properties file can't be found
	 * @throws IOException if the properties file can't be read
	 */
	protected BasicRequest() 
	throws FileNotFoundException, IOException {
		Properties props = Utils.getAppProps();
		this.httpsHost = "https://" + props.getProperty("application.host") 
			+ ":" + props.getProperty("application.sslport");
		this.httpHost  = "http://" + props.getProperty("application.host") 
			+ ":" + props.getProperty("application.port");
	}
	
	/**
	 * Create a BasicRequest with a url and context path
	 * 
	 * @param url			The requested url
	 * @param contextPath	The context path, e.g. "/loband"
	 * @throws Exception	If it can't read the properties file
	 */
    public BasicRequest(URL url, String contextPath)
    throws FileNotFoundException, IOException
	{
    	this();
        this.setURL(url);
        this.contextPath = contextPath;
    }
    
    /**
     * Create a BasicRequest with a url and a servlet request
     * 
     * @param url				The requested url
     * @param servletRequest	The servlet request
     * @throws Exception If it can't read the properties file	
     */
    public BasicRequest(URL url, HttpServletRequest servletRequest) 
    throws FileNotFoundException, IOException
	{
    	this(servletRequest);
    	this.setURL(url);
    }
    
    /**
     * Create a BasicRequest with a servlet request
     * 
     * @param servletRequest 	The servlet request
     * @throws Exception If it can't read the properties file
     */
    public BasicRequest(HttpServletRequest servletRequest) 
    throws FileNotFoundException, IOException
    {
    	this();
        this.contextPath = servletRequest.getContextPath();
    }
    
    /**
     * This method returns the headers of a request.
     * 
     * @return The headers
     */
	public Map getHeaders() {
		return headers;
	}

	/**
	 * Set the headers for the request
	 * 
	 * @param headers The http headers
	 */
	public void setHeaders(Map headers) {
		this.headers = headers;
	}

	/**
	 * Returns a parameter of the request
	 * 
	 * @param name The name of the parameter
	 * @return The parameter
	 */
	public Object getParameter(String name) {
		return getParameters().get(name);
	}

	/**
	 * Set a parameter for the request
	 * 
	 * @param name		The name of the parameter
	 * @param values	The values of the parameter
	 * @return All the parameters for the request
	 */
	public Object setParameter(String name, String [] values) {
		return getParameters().put(name, values);
	}

	/**
	 * Returns the parameters for the request
	 * 
	 * @return The parameters
	 */
	public Map getParameters() {
		return this.parameters;
	}

	/**
	 * Set all the parameters for the request
	 * 
	 * @param parameters All the parameters
	 */
	public void setParameters(Map parameters) {
		this.parameters = parameters;
	}
	
	/**
	 * Returns the url of the request
	 * 
	 * @return The url
	 */
	public URL getURL() {
        return url;
	}
	
	/**
	 * Sets the url for the request
	 * 
	 * @param url The url
	 */
	public void setURL(URL url) {
        this.url = url;
	}
    
    
    /**
     * Returns the context path, for instance the root part of the original
     * URL that is required for appending to the begining of all rewritten
     * links.
     * 
     * @return Theo context path
     */
    public String getContextPath() {
        return contextPath;
    }
    
    /**
     * This method increases the recursion depth for redirects
     */
    public void increaseDepth() throws RecursionDepthException {
        depth++;
        if (depth > maxDepth) {
            throw new RecursionDepthException("recursion depth exceeded "+maxDepth);
        }
    }
    
    /**
     * This method decreases the recursion depth for redirects
     */
    public void decreaseDepth() throws RecursionNestingException {
        depth--;
        if (depth < 0) {
            throw new RecursionNestingException("recursion depth went below 0, "+
                "too many calls to decreaseDepth");
        }
    }
    
    /**
     * Returns the recursion depth
     * 
     * @return The recursion depth 
     */
    public int getDepth() {
        return depth;
    }

	/**
	 * This method sets the Http state for the request
	 * 
	 * @param state The Http state
	 */
    public void setState(HttpState state){
    	this.state = state;
    }
    
    /**
     * Returns the Http state for the request
     * 
     * @return The Http state
     */
    public HttpState getState(){
    	return state;
    }
    
    /**
     * This function sets the http method for the request
     * 
     * @param method The GET/POST method 
     */
    public void setMethod(Method method) {
    	this.method = method;
    }
    
    /**
     * Returns the http method for the request
     * 
     * @return The http Method
     */
    public Method getMethod() {
    	return method;
    }

    /**
     * Will return all the PartInfo's of the uploaded files
     * 
     * @return The PartInfo's
     */
    public PartInfo[] getParts(){
    	return parts;
    }
    
    /**
     * Set all the uploaded files for the request
     * 
     * @param files The files
     */
    public void setParts(PartInfo[] files){
    	parts = files;
    }
    
    /**
     * Returns the timer that keeps track of how long 
     * the request takes.
     * 
     * @return The timer
     */
    public Timer getTimer() {
    	return timer;
    }
	
    /**
     * Set the filter parameters
     * 
     * @param params The parameters
     */
	public void setFilterParams(Map params) {
		this.filter_params = params;
	}
	
	/**
	 * Returns the filter parameters
	 * 
	 * @return The parameters
	 */
	public Map getFilterParams() {
		return this.filter_params;
	}

	/**
	 * Returns the secure url of the host 
	 * 
	 * @return Secure host Url 
	 */
	public String getHttpsHost() {
		return httpsHost;
	}
	
	/**
	 * Returns the url of the host
	 * 
	 * @return Host Url
	 */
	public String getHttpHost() {
		return httpHost;
	}
	
	/*
	 * Get / Set Simulator speed in bytes per second
	 * 
	 */
	public int getSimulatorBps() {
		return simulatorBps;
	}
	
	public void setSimulatorBps(int bps) {
		simulatorBps = bps;
	}

	/*
	 * Get / Set Simulator key - all requests with the same key are
	 *  throttled such that their total bandwidth is as set by 
	 * setSimulatorBps 
	 */

	public int getSimulatorKey() {
		return simulatorKey;
	}
	
	public void setSimulatorKey(int key) {
		simulatorKey = key;
	}

	/*
	 * Get / Set simulator on or off. true = on, false = off
	 * 
	 */

	public boolean getSimulatorOnOff() {
		return simulatorOnOff;
	}
	
	public void setSimulatorOnOff(boolean on) {
		simulatorOnOff = on;
	}
	
	public String getRequestType() {
		return requestType;
	}
	
	public void setRequestType(String s) {
		requestType=s;
	}
	
	public void setFiltering(boolean x) {
		bFiltering=x;
	}
	public boolean getFiltering() {
		return(bFiltering);
	}
	protected void copyImpl(BasicRequest newReq) {
		newReq.headers        = new Hashtable(headers);
		newReq.depth          = depth;
		newReq.contextPath    = contextPath;
		newReq.url            = url;
		newReq.method         = method;
		newReq.parameters     = new Hashtable(parameters);
		newReq.filter_params  = new Hashtable(filter_params);
		newReq.httpsHost      = httpsHost;
		newReq.httpHost       = httpHost;
		newReq.simulatorBps	  = simulatorBps;
		newReq.simulatorKey	  = simulatorKey;
		newReq.simulatorOnOff = simulatorOnOff;
		newReq.requestType    = requestType;
		newReq.bFiltering     = bFiltering;
	    newReq.timer = (Timer)( this.timer.clone() );
	    if (this.state != null)
	    {
			newReq.state = new HttpState();
			newReq.state.addCookies(this.state.getCookies());
			// TODO/FIXME anything else to copy in the state?
	    }
	    if (this.parts != null)
	    {
	    	newReq.parts = new PartInfo[this.parts.length];
	    	System.arraycopy(this.parts, 0, newReq.parts, 0, this.parts.length);
	    	// TODO/FIXME are PartInfos mutable? If so we should clone them
	    }
	}
	
    /**
     * Returns the array of query parameter names 
     * <i>in the order they appear in the url</i> - the usual methods for 
     * getting parameter names do not
     * preserve this order, and can cause pages not to be found.
     * @param queryString  The request query string to parse
     * @return The array of parameter names in the order they appear in the url
     */    
    public String[] getOrderedParamNames() {
        return (String[])(parameters.keySet().toArray(new String[]{}) );
    }
}
