/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */

package org.aidworld.aidbase.request;

import java.io.ByteArrayOutputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Map;

import javax.servlet.ServletInputStream;
import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServletRequest;

import org.aidworld.aidbase.error.FormDataDecodeException;
import org.aidworld.aidbase.util.Codecs;
import org.aidworld.aidbase.util.NVPair;
import org.aidworld.aidbase.util.PartInfo;
import org.aidworld.aidbase.util.Utils;
import org.apache.log4j.Logger;

/**
 * This type of Request represents a request that came in by HTTP,
 * specifically a Java Servlet engine of some kind. It requires a
 * ServletRequest object which it stores locally.
 */
public class HttpRequest extends BasicRequest {
	private ServletRequest request;
    private static Logger LOG = Logger.getLogger(HttpRequest.class);
    
	private HttpRequest ()
	{
        super(true);
	}
	
    /**
	 * Constructor which takes a http servlet request
	 * 
	 * @param request	 The http servlet request
	 * @throws FileNotFoundException if the application properties file can't be found
	 * @throws FormDataDecodeException if the POST request contained multipart form data that could not be decoded
	 * @throws IOException if the application properties file can't be read, or reading POST form data from the client failed
	 * @throws IllegalArgumentException if a header is duplicated
	 */
	public HttpRequest (HttpServletRequest request) 
	throws FileNotFoundException, FormDataDecodeException, 
	IOException, IllegalArgumentException 
	{
        super(request);
		this.request = request;
        
        Hashtable headers = new Hashtable();
                
        for (Enumeration names = request.getHeaderNames(); names.hasMoreElements(); ) {
            String name = (String)(names.nextElement());
            if (headers.get(name) != null) {
                throw new IllegalArgumentException("repeated header: "+name);
            }
            String value = request.getHeader(name);  
            headers.put(name, value);
        }

        setHeaders(headers);
        setParameters(request.getParameterMap());
        
		if (request.getMethod().equals("POST") && 
			request.getContentType().startsWith("multipart/form-data")) 
		{ 
			getTimer().push("reading POST form data");
			decodeMultipartUpload();
	        getTimer().pop();
		}
	}

	/**
	 * Constuctor which takes an url and a http servlet request
	 * 
	 * @param url			The url of the requested page
	 * @param request		The http servlet request
	 * @throws Exception    If it has dubble headers or the request is not a http servlet request
	 */
	public HttpRequest(URL url, HttpServletRequest request) throws Exception
	{
		this(request);
		this.setURL(url);
	}
	
	/**
	 * Returns the servlet request
	 * 
	 * @return The servlet request
	 */
	public ServletRequest getRequest() {
		return request;
	}
	
	private void decodeMultipartUpload() 
	throws IOException, FormDataDecodeException 
	{
    	Map params = new Hashtable();
    	NVPair[] parts = null;
    	ByteArrayOutputStream byteBuffer = new ByteArrayOutputStream();
        byte[] buffer = new byte[4096];
        ServletInputStream input = request.getInputStream();
       
		while (true) {
			LOG.debug("reading block");
			int bytesRead = input.read(buffer);
			if (bytesRead == -1) break;
			LOG.debug("writing block: "+bytesRead+" bytes");
			byteBuffer.write(buffer, 0, bytesRead);
		}
        
        try {
   	 		NVPair[][] decoded = Codecs.mpFormDataDecode(
   	 				byteBuffer.toByteArray(),
   	 				request.getContentType(),
					Utils.getTempDirectory());

   	 		NVPair[] nvParams = decoded[0];
   	 		// NVPair[] nvFiles  = decoded[1];
   	 		parts 			  = decoded[2];
   	 		
   	 		for (int i=0;i<nvParams.length;i++){
   	 			// System.out.println(nvParams[i].getName() +" = "+ nvParams[i].getValue());
   	 			params.put(nvParams[i].getName(),new String[]{nvParams[i].getValue()});
			}
   	 		
   	 		Utils.dumpFile("/tmp/FormData-recieved", byteBuffer.toByteArray());
   	 		this.setParts((PartInfo[])( parts ));
   	 		this.setParameters(params);
        } 
        catch (Exception p)
		{
        	throw new FormDataDecodeException(
        			"Error decoding form data: "+p, p);
        }   	 	
	}
	
	public Request copy() {
		HttpRequest req = new HttpRequest();
		super.copyImpl(req);
		req.request = this.request;
		return req;
	}
	
    /**
     * Returns the array of query parameter names 
     * <i>in the order they appear in the url</i> - the usual methods for 
     * getting parameter names do not
     * preserve this order, and can cause pages not to be found.
     * @param queryString  The request query string to parse
     * @return The array of parameter names in the order they appear in the url
     */    
    public String[] getOrderedParamNames() {
        if (!(request instanceof HttpServletRequest)) 
        {
            // If its not an HttpServletRequest then we don't have access
        	// to the original query string, so we have to regard all
        	// parameters as GET parameters (for now).
            return super.getOrderedParamNames(); 
        }

    	HttpServletRequest sr = (HttpServletRequest)(request); 
        String queryString = sr.getQueryString();

        if (queryString == null) return new String[0];
        String[] params = queryString.split("&");
        
        for (int i = 0; i < params.length; i++){
            String s = params[i];
            if (s.indexOf("=") != -1){
                params[i] = s.substring(0, s.indexOf("="));
            }
        }
        return params;
    }


}
