/*
 * @(#)Codecs.java					0.3-3 06/05/2001
 *
 *  This file is part of the HTTPClient package
 *  Copyright (C) 1996-2001 Ronald Tschal
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307, USA
 *
 *  For questions, suggestions, bug-reports, enhancement-requests etc.
 *  I may be contacted at:
 *
 *  ronald@innovation.ch
 *
 *  The HTTPClient's home page is located at:
 *
 *  http://www.innovation.ch/java/HTTPClient/ 
 *
 */

package org.aidworld.aidbase.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLConnection;
import java.util.BitSet;
import java.util.Vector;

import org.apache.log4j.Logger;


/**
 * This class collects various encoders and decoders.
 *
 * @version	0.3-3  06/05/2001
 * @author	Ronald Tschal
 */
public class Codecs
{
	private static BitSet  BoundChar;
	private static BitSet  EBCDICUnsafeChar;
    private static byte[]  Base64EncMap, Base64DecMap;
    private static char[]  UUEncMap;
    private static byte[]  UUDecMap;

	
	private final static String ContDisp = "\r\nContent-Disposition: form-data; name=\"";
	private final static String FileName = "\"; filename=\"";
	private final static String ContType = "\r\nContent-Type: ";
	private final static String Boundary = "\r\n----------ieoau._._+2_8_GoodLuck8.3-dskdfJwSJKl234324jfLdsjfdAuaoei-----";
	
	private static Logger LOG = Logger.getLogger(Codecs.class);
    
	// Class Initializer
	
	static
	{
		// rfc-2046 & rfc-2045: (bcharsnospace & token)
		// used for multipart codings
		BoundChar = new BitSet(256);
		for (int ch='0'; ch <= '9'; ch++)  BoundChar.set(ch);
		for (int ch='A'; ch <= 'Z'; ch++)  BoundChar.set(ch);
		for (int ch='a'; ch <= 'z'; ch++)  BoundChar.set(ch);
		BoundChar.set('+');
		BoundChar.set('_');
		BoundChar.set('-');
		BoundChar.set('.');
		
		// EBCDIC unsafe characters to be quoted in quoted-printable
		// See first NOTE in section 6.7 of rfc-2045
		EBCDICUnsafeChar = new BitSet(256);
		EBCDICUnsafeChar.set('!');
		EBCDICUnsafeChar.set('"');
		EBCDICUnsafeChar.set('#');
		EBCDICUnsafeChar.set('$');
		EBCDICUnsafeChar.set('@');
		EBCDICUnsafeChar.set('[');
		EBCDICUnsafeChar.set('\\');
		EBCDICUnsafeChar.set(']');
		EBCDICUnsafeChar.set('^');
		EBCDICUnsafeChar.set('`');
		EBCDICUnsafeChar.set('{');
		EBCDICUnsafeChar.set('|');
		EBCDICUnsafeChar.set('}');
		EBCDICUnsafeChar.set('~');
		
		
        // rfc-2045: Base64 Alphabet
        byte[] map = {
            (byte)'A', (byte)'B', (byte)'C', (byte)'D', (byte)'E', (byte)'F',
            (byte)'G', (byte)'H', (byte)'I', (byte)'J', (byte)'K', (byte)'L',
            (byte)'M', (byte)'N', (byte)'O', (byte)'P', (byte)'Q', (byte)'R',
            (byte)'S', (byte)'T', (byte)'U', (byte)'V', (byte)'W', (byte)'X',
            (byte)'Y', (byte)'Z',
            (byte)'a', (byte)'b', (byte)'c', (byte)'d', (byte)'e', (byte)'f',
            (byte)'g', (byte)'h', (byte)'i', (byte)'j', (byte)'k', (byte)'l',
            (byte)'m', (byte)'n', (byte)'o', (byte)'p', (byte)'q', (byte)'r',
            (byte)'s', (byte)'t', (byte)'u', (byte)'v', (byte)'w', (byte)'x',
            (byte)'y', (byte)'z',
            (byte)'0', (byte)'1', (byte)'2', (byte)'3', (byte)'4', (byte)'5',
            (byte)'6', (byte)'7', (byte)'8', (byte)'9', (byte)'+', (byte)'/' };
        Base64EncMap = map;
        Base64DecMap = new byte[128];
        for (int idx=0; idx<Base64EncMap.length; idx++)
            Base64DecMap[Base64EncMap[idx]] = (byte) idx;

        // uuencode'ing maps
        UUEncMap = new char[64];
        for (int idx=0; idx<UUEncMap.length; idx++)
            UUEncMap[idx] = (char) (idx + 0x20);
        UUDecMap = new byte[128];
        for (int idx=0; idx<UUEncMap.length; idx++)
            UUDecMap[UUEncMap[idx]] = (byte) idx;
    }

	// Constructors
	
	/**
	 * This class isn't meant to be instantiated.
	 */
	private Codecs() {}
	
	
	// Methods
	  /**
     * This method decodes the given string using the base64-encoding
     * specified in RFC-2045 (Section 6.8).
     *
     * @param  str the base64-encoded string.
     * @return the decoded <var>str</var>.
     */
    public final static String base64Decode(String str)
    {
        if (str == null)  return  null;

        try
            { return new String(base64Decode(str.getBytes("8859_1")), "8859_1"); }
        catch (UnsupportedEncodingException uee)
            { throw new Error(uee.toString()); }
    }

    public final static byte[] base64Decode(byte[] data)
    {
        if (data == null)  return  null;

        int tail = data.length;
        while (data[tail-1] == '=')  tail--;

        byte dest[] = new byte[tail - data.length/4];


        // ascii printable to 0-63 conversion
        for (int idx = 0; idx <data.length; idx++)
            data[idx] = Base64DecMap[data[idx]];

        // 4-byte to 3-byte conversion
        int sidx, didx;
        for (sidx = 0, didx=0; didx < dest.length-2; sidx += 4, didx += 3)
        {
            dest[didx]   = (byte) ( ((data[sidx] << 2) & 255) |
                            ((data[sidx+1] >>> 4) & 003) );
            dest[didx+1] = (byte) ( ((data[sidx+1] << 4) & 255) |
                            ((data[sidx+2] >>> 2) & 017) );
            dest[didx+2] = (byte) ( ((data[sidx+2] << 6) & 255) |
                            (data[sidx+3] & 077) );
        }
        if (didx < dest.length)
            dest[didx]   = (byte) ( ((data[sidx] << 2) & 255) |
                            ((data[sidx+1] >>> 4) & 003) );
        if (++didx < dest.length)
            dest[didx]   = (byte) ( ((data[sidx+1] << 4) & 255) |
                            ((data[sidx+2] >>> 2) & 017) );

        return dest;
    }

	
	
	/**
	 * This method decodes a multipart/form-data encoded string. The boundary
	 * is parsed from the <var>cont_type</var> parameter, which must be of the
	 * form 'multipart/form-data; boundary=...'. Any encoded files are created
	 * in the directory specified by <var>dir</var> using the encoded filename.
	 *
	 * <P><em>Note:</em> Does not handle nested encodings (yet).
	 *
	 * <P>Examples: If you're receiving a multipart/form-data encoded response
	 * from a server you could use something like:
	 * <PRE>
	 *     NVPair[] opts = Codecs.mpFormDataDecode(resp.getData(),
	 *                                  resp.getHeader("Content-type"), ".");
	 * </PRE>
	 * If you're using this in a Servlet to decode the body of a request from
	 * a client you could use something like:
	 * <PRE>
	 *     byte[] body = new byte[req.getContentLength()];
	 *     new DataInputStream(req.getInputStream()).readFully(body);
	 *     NVPair[] opts = Codecs.mpFormDataDecode(body, req.getContentType(),
	 *                                             ".");
	 * </PRE>
	 * (where 'req' is the HttpServletRequest).
	 *
	 * <P>Assuming the data received looked something like:
	 * <PRE>
	 * -----------------------------114975832116442893661388290519
	 * Content-Disposition: form-data; name="option"
	 *                                                         &nbsp;
	 * doit
	 * -----------------------------114975832116442893661388290519
	 * Content-Disposition: form-data; name="comment"; filename="comment.txt"
	 *                                                         &nbsp;
	 * Gnus and Gnats are not Gnomes.
	 * -----------------------------114975832116442893661388290519--
	 * </PRE>
	 * you would get one file called <VAR>comment.txt</VAR> in the current
	 * directory, and opts would contain two elements: {"option", "doit"}
	 * and {"comment", "comment.txt"}
	 *
	 * @param     data        the form-data to decode.
	 * @param     cont_type   the content type header (must contain the
	 *			      boundary string).
	 * @param     dir         the directory to create the files in.
	 * @param     mangler     the filename mangler, or null if no mangling is
	 *                        to be done. This is invoked just before each
	 *                        file is created and written, thereby allowing
	 *                        you to control the names of the files.
	 * @return                an array of name/value pairs, one for each part;
	 *                        the name is the 'name' attribute given in the
	 *                        Content-Disposition header; the value is either
	 *                        the name of the file if a filename attribute was
	 *                        found, or the contents of the part.
	 * @exception IOException If any file operation fails.
	 * @exception ParseException If an error during parsing occurs.
	 */
	public final static NVPair[][] mpFormDataDecode(byte[] data, String cont_type,
			String dir)
	throws IOException, ParseException
	{
		// Find and extract boundary string
		
		String bndstr = HttpClientUtils.getParameter("boundary", cont_type);
		if (bndstr == null)
			throw new ParseException("'boundary' parameter not found in Content-type: " + cont_type);
		
		byte[] srtbndry = (    "--" + bndstr + "\r\n").getBytes("8859_1"),
		boundary = ("\r\n--" + bndstr + "\r\n").getBytes("8859_1"),
		endbndry = ("\r\n--" + bndstr + "--"  ).getBytes("8859_1");
		
		
		// setup search routines
		
		int[] bs = HttpClientUtils.compile_search(srtbndry),
		bc = HttpClientUtils.compile_search(boundary),
		be = HttpClientUtils.compile_search(endbndry);
		
		
		// let's start parsing the actual data
		
		int start = HttpClientUtils.findStr(srtbndry, bs, data, 0, data.length);
		if (start == -1)	// didn't even find the start
			throw new ParseException("Starting boundary not found: " +
					new String(srtbndry, "8859_1"));
		start += srtbndry.length;
		
		Vector params  = new Vector();
		Vector files   = new Vector();
		Vector parts   = new Vector();
		boolean  done  = false;
		int      idx;
		
		for (idx=0; !done; idx++)
		{
			// find end of this part
			
			int end = HttpClientUtils.findStr(boundary, bc, data, start, data.length);
			if (end == -1)		// must be the last part
			{
				end = HttpClientUtils.findStr(endbndry, be, data, start, data.length);
				if (end == -1)
					throw new ParseException("Ending boundary not found: " +
							new String(endbndry, "8859_1"));
				done = true;
			}
			
			// parse header(s)
			
			String hdr, name = null, filename = null, cont_disp = null;
			
			while (true)
			{
				int next = findEOL(data, start) + 2;
				if (next-2 <= start)  break;	// empty line -> end of headers
				hdr      = new String(data, start, next-2-start, "8859_1");
				start    = next;
				
				// handle line continuation
				byte ch;
				while (next < data.length-1  &&
						((ch = data[next]) == ' '  ||  ch == '\t'))
				{
					next   = findEOL(data, start) + 2;
					hdr   += new String(data, start, next-2-start, "8859_1");
					start  = next;
				}
				
				if (!hdr.regionMatches(true, 0, "Content-Disposition", 0, 19))
					continue;
				Vector pcd =
					HttpClientUtils.parseHeader(hdr.substring(hdr.indexOf(':')+1));
				HttpHeaderElement elem = HttpClientUtils.getElement(pcd, "form-data");
				Vector temp =HttpClientUtils.parseHeader(elem.toString());
				
				for (int i=0;i<temp.size();i++){
					LOG.debug(temp.get(i).toString());
				}
				
				if (elem == null)
					throw new ParseException("Expected 'Content-Disposition: form-data' in line: "+hdr);
				
				NVPair[] elemParams = elem.getParams();
				name = filename = null;
				for (int pidx=0; pidx<elemParams.length; pidx++)
				{
					if (elemParams[pidx].getName().equalsIgnoreCase("name"))
						name = elemParams[pidx].getValue();
					if (elemParams[pidx].getName().equalsIgnoreCase("filename"))
						filename = elemParams[pidx].getValue();
				}
				if (name == null)
					throw new ParseException("'name' parameter not found in header: "+hdr);
				
				cont_disp = hdr;
			}
			
			start += 2;
			if (start > end)
				throw new ParseException("End of header not found at offset "+end);
			
			if (cont_disp == null)
				throw new ParseException("Missing 'Content-Disposition' header at offset "+start);
			
			// handle data for this part
			String randomName = "Upload_" + Math.random();
			String originalName = "";
			if (filename != null)			// It's a file
			{
				if (filename != null  &&  filename.length() > 0)
				{
					originalName = filename;
					File file = new File(dir, randomName);
					FileOutputStream out = new FileOutputStream(file);
					filename = file.getAbsolutePath();
					out.write(data, start, end-start);
					out.close();
				}
				files.add(new NVPair(name,originalName));
				parts.add(new FileInfo(name,originalName,randomName));
			}
			else					// It's simple data
			{
				params.add(new NVPair(name,new String(data, start, end-start, "8859_1")));
				parts.add(new ParamInfo(name,new String(data, start, end-start, "8859_1")));
			}
				
			start = end + boundary.length;
		}
		NVPair[] nvParams  = (NVPair[])( params.toArray(new NVPair[0]) );
		NVPair[] nvFiles   = (NVPair[])( files.toArray(new NVPair[0]) );
		PartInfo[] nvParts = (PartInfo[])( parts.toArray(new PartInfo[0]) ); 
		
		
		return new NVPair[][]{nvParams,nvFiles,nvParts};
	}
	
	
	/**
	 * This method encodes name/value pairs and files into a byte array
	 * using the multipart/form-data encoding. The boundary is returned
	 * as part of <var>ct_hdr</var>.
	 * <BR>Example:
	 * <PRE>
	 *     NVPair[] opts = { new NVPair("option", "doit") };
	 *     NVPair[] file = { new NVPair("comment", "comment.txt") };
	 *     NVPair[] hdrs = new NVPair[1];
	 *     byte[]   data = Codecs.mpFormDataEncode(opts, file, hdrs);
	 *     con.Post("/cgi-bin/handle-it", data, hdrs);
	 * </PRE>
	 * <VAR>data</VAR> will look something like the following:
	 * <PRE>
	 * -----------------------------114975832116442893661388290519
	 * Content-Disposition: form-data; name="option"
	 *                                                         &nbsp;
	 * doit
	 * -----------------------------114975832116442893661388290519
	 * Content-Disposition: form-data; name="comment"; filename="comment.txt"
	 * Content-Type: text/plain
	 *                                                         &nbsp;
	 * Gnus and Gnats are not Gnomes.
	 * -----------------------------114975832116442893661388290519--
	 * </PRE>
	 * where the "Gnus and Gnats ..." is the contents of the file
	 * <VAR>comment.txt</VAR> in the current directory.
	 *
	 * <P>If no elements are found in the parameters then a zero-length
	 * byte[] is returned and the content-type is set to
	 * <var>application/octet-string</var> (because a multipart must
	 * always have at least one part.
	 *
	 * <P>For files an attempt is made to discover the content-type, and if
	 * found a Content-Type header will be added to that part. The content type
	 * is retrieved using java.net.URLConnection.guessContentTypeFromName() -
	 * see java.net.URLConnection.setFileNameMap() for how to modify that map.
	 * Note that under JDK 1.1 by default the map seems to be empty. If you
	 * experience troubles getting the server to accept the data then make
	 * sure the fileNameMap is returning a content-type for each file (this
	 * may mean you'll have to set your own).
	 *
	 * @param     opts        the simple form-data to encode (may be null);
	 *                        for each NVPair the name refers to the 'name'
	 *                        attribute to be used in the header of the part,
	 *                        and the value is contents of the part.
	 *                        null elements in the array are ingored.
	 * @param     files       the files to encode (may be null); for each
	 *                        NVPair the name refers to the 'name' attribute
	 *                        to be used in the header of the part, and the
	 *                        value is the actual filename (the file will be
	 *                        read and it's contents put in the body of
	 *                        that part). null elements in the array
	 *                        are ingored.
	 * @param     ct_hdr      this returns a new NVPair in the 0'th element
	 *                        which contains name = "Content-Type",
	 *			      value = "multipart/form-data; boundary=..."
	 *                        (the reason this parameter is an array is
	 *                        because a) that's the only way to simulate
	 *                        pass-by-reference and b) you need an array for
	 *                        the headers parameter to the Post() or Put()
	 *                        anyway). The exception to this is that if no
	 *                        opts or files are given the type is set to
	 *                        "application/octet-stream" instead.
	 * @param     mangler     the filename mangler, or null if no mangling is
	 *                        to be done. This allows you to change the name
	 *                        used in the <var>filename</var> attribute of the
	 *                        Content-Disposition header. Note: the mangler
	 *                        will be invoked twice for each filename.
	 * @return                an encoded byte array containing all the opts
	 *			      and files.
	 * @exception IOException If any file operation fails.
	 */
	public final static byte[] mpFormDataEncode(PartInfo[] parts,
			NVPair[] ct_hdr)
	throws IOException
	{
		byte[] boundary  = Boundary.getBytes("8859_1"),
		cont_disp = ContDisp.getBytes("8859_1"),
		cont_type = ContType.getBytes("8859_1"),
		filename  = FileName.getBytes("8859_1");
		int len = 0,
		hdr_len = boundary.length + cont_disp.length+1 + 2 +  2;
		//        \r\n --  bnd      \r\n C-D: ..; n=".." \r\n \r\n
	
		
		// Calculate the length of the data
		
		for (int idx=0; idx<parts.length; idx++)
		{
			PartInfo part = parts[idx];
			if (part == null)  continue;
			if (part instanceof ParamInfo){
				len += hdr_len + part.getName().length() +
				part.getValue().length();
			}else if (part instanceof FileInfo){
				FileInfo fi = (FileInfo) ( part );
				File file = new File(Utils.getTempDirectory()+
							"/"+fi.getRandom());
				String fname = fi.getValue();
				if (fname != null)
				{
					len += hdr_len + part.getName().length() + filename.length;
					len += fname.length() + file.length();
					
					String ct = CT.getContentType(file.getName());
					if (ct == null && fname.equals("")){
						ct = "application/octet-stream";
					}
					if (ct != null)
						len += cont_type.length + ct.length();
				}
			
			}
			
		}
		
		if (len == 0)
		{
			ct_hdr[0] = new NVPair("Content-Type", "application/octet-stream");
			return new byte[0];
		}
		
		len -= 2;			// first CR LF is not written
		len += boundary.length + 2 + 2;	// \r\n -- bnd -- \r\n
		
		
		// Now fill array
		
		byte[] res = new byte[len];
		int    pos = 0;
		
		NewBound: for (int new_c=0x30303030; new_c!=0x7A7A7A7A; new_c++)
		{
			pos = 0;
			
			// modify boundary in hopes that it will be unique
			while (!BoundChar.get(new_c     & 0xff)) new_c += 0x00000001;
			while (!BoundChar.get(new_c>>8  & 0xff)) new_c += 0x00000100;
			while (!BoundChar.get(new_c>>16 & 0xff)) new_c += 0x00010000;
			while (!BoundChar.get(new_c>>24 & 0xff)) new_c += 0x01000000;
			boundary[40] = (byte) (new_c     & 0xff);
			boundary[42] = (byte) (new_c>>8  & 0xff);
			boundary[44] = (byte) (new_c>>16 & 0xff);
			boundary[46] = (byte) (new_c>>24 & 0xff);
			
			int off = 2;
			int[] bnd_cmp = HttpClientUtils.compile_search(boundary);
			
			for (int idx=0; idx<parts.length; idx++)
			{
				PartInfo part = parts[idx];
				if (part == null)  continue;

				
				System.arraycopy(boundary, off, res, pos, boundary.length-off);
				pos += boundary.length - off;
				off  = 0;
				int  start = pos;
				
				System.arraycopy(cont_disp, 0, res, pos, cont_disp.length);
				pos += cont_disp.length;
				
				int nlen = part.getName().length();
				System.arraycopy(part.getName().getBytes("8859_1"), 0, res, pos, nlen);
				pos += nlen;
				File file=null;
				
				if(part instanceof FileInfo){
					FileInfo fi = (FileInfo) ( part );
					file = new File(Utils.getTempDirectory()+
								"/"+fi.getRandom());
					String fname = fi.getValue();
					
					if (fname == null)  continue;
					
					System.arraycopy(filename, 0, res, pos, filename.length);
					pos += filename.length;
					
					nlen = fname.length();
					System.arraycopy(fname.getBytes("8859_1"), 0, res, pos, nlen);
					pos += nlen;
					
					res[pos++] = (byte) '"';
					
					String ct = CT.getContentType(file.getName());
					if (ct == null && fname.equals("")){
						ct = "application/octet-stream";
					}
					
					if (ct != null)
					{
						System.arraycopy(cont_type, 0, res, pos, cont_type.length);
						pos += cont_type.length;
						System.arraycopy(ct.getBytes("8859_1"), 0, res, pos, ct.length());
						pos += ct.length();
					}
				}else if(part instanceof ParamInfo){
					res[pos++] = (byte) '"';
				}
				
				res[pos++] = (byte) '\r';
				res[pos++] = (byte) '\n';
				res[pos++] = (byte) '\r';
				res[pos++] = (byte) '\n';
				
				if (part instanceof ParamInfo){
					int vlen = part.getValue().length();
					System.arraycopy(part.getValue().getBytes("8859_1"), 0, res, pos, vlen);
					pos += vlen;
				} else if (part instanceof FileInfo && !part.getValue().equals("")){
					nlen = (int) file.length();
					FileInputStream fin = new FileInputStream(file);
					while (nlen > 0)
					{
						int got = fin.read(res, pos, nlen);
						nlen -= got;
						pos += got;
					}
					fin.close();
					
				}
				
				
				if ((pos-start) >= boundary.length  &&
						HttpClientUtils.findStr(boundary, bnd_cmp, res, start, pos) != -1)
					continue NewBound;
			}
			
			break NewBound;
		}
		
		System.arraycopy(boundary, 0, res, pos, boundary.length);
		pos += boundary.length;
		res[pos++] = (byte) '-';
		res[pos++] = (byte) '-';
		res[pos++] = (byte) '\r';
		res[pos++] = (byte) '\n';
		
		if (pos != len)
			throw new Error("Calculated "+len+" bytes but wrote "+pos+" bytes!");
		
		/* the boundary parameter should be quoted (rfc-2046, section 5.1.1)
		 * but too many script authors are not capable of reading specs...
		 * So, I give up and don't quote it.
		 */
		ct_hdr[0] = new NVPair("Content-Type",
				"multipart/form-data; boundary=" +
				new String(boundary, 4, boundary.length-4, "8859_1"));
		
		return res;
	}
	
	private static class CT extends URLConnection
	{
		protected static final String getContentType(String fname)
		{
			return guessContentTypeFromName(fname);
		}
		
		private CT() { super(null); }
		public void connect() { }
	}
	
	/**
	 * Searches for the next CRLF in an array.
	 *
	 * @param  arr the byte array to search.
	 * @param  off the offset at which to start the search.
	 * @return the position of the CR or (arr.length-2) if not found
	 */
	private final static int findEOL(byte[] arr, int off)
	{
		while (off < arr.length-1  &&
				!(arr[off++] == '\r'  &&  arr[off] == '\n'));
		return off-1;
	}
	
	
	
}
