/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */

package org.aidworld.aidbase.webui.servlet;

import java.io.IOException;
import java.net.URL;
import java.sql.SQLException;

import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.aidworld.aidbase.error.BadOutputException;
import org.aidworld.aidbase.error.BlockedUrlException;
import org.aidworld.aidbase.error.DocumentNotFoundException;
import org.aidworld.aidbase.error.GoDirectException;
import org.aidworld.aidbase.error.IncompatibleDocumentException;
import org.aidworld.aidbase.error.RecursionDepthException;
import org.aidworld.aidbase.error.RecursionNestingException;
import org.aidworld.aidbase.error.RequestException;
import org.aidworld.aidbase.error.ServerProtocolException;
import org.aidworld.aidbase.error.ServerTimeoutException;
import org.aidworld.aidbase.webui.util.JSPManager;
import org.aidworld.aidbase.webui.util.WebUtils;
import org.apache.log4j.Logger;

/**
 * Base class for Aidbase servlets.
 * This class is based on one written by MIT for the DSpace project.
 */

public class AidbaseServlet extends HttpServlet {

    /* log4j category */
    private static Logger LOG = Logger.getLogger(AidbaseServlet.class);
    
    // Because the error handling is indentical for GET and POST requests,
    // to make things easier, doGet and doPost are folded into one method
    // which then "re-separates" to GET and POST.

    /** Overrides doGet in HttpServlet to delegate to processRequest */
    public final void doGet(HttpServletRequest request, HttpServletResponse response)
        throws ServletException, IOException
    {
        try {
        	doABGet(request, response);
        } catch (Exception e) {
            // Log the exception and show a suitably apologetic error
            LOG.error("Something went terribly wrong!!", e); // I18N 3000
            JSPManager.showInternalError(request, response, e);
        }
    }

    /** Overrides doPost in HttpServlet to delegate to processRequest */
    public final void doPost(HttpServletRequest request,
        HttpServletResponse response)
        throws ServletException, IOException
    {
        try {
        	doABPost(request, response);
        } catch (Exception e) {
            // Log the exception and show a suitably apologetic error
            LOG.error("Something went terribly wrong!!", e); // I18N 3000
            JSPManager.showInternalError(request, response, e);
        }
    }

    /**
     * Process an incoming HTTP GET. Override this method instead of doGet().
     *
     * @param request    the HTTP request
     * @param response   the HTTP response
     *
     * @throws SQLException  if a database error occurs
     * @throws AuthorizeException  if some authorization error occurs
     */
    protected void doABGet(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException
    {
        // If this is not overridden, we invoke the raw HttpServlet "doGet" to
        // indicate that GET is not supported by this servlet.
        super.doGet(request, response);
    }

    /**
     * Process an incoming HTTP POST. Override this method instead of doPost().
     *
     * @param request    the HTTP request
     * @param response   the HTTP response
     *
     * @throws ServletException if the 
     * @throws AuthorizeException  if some authorization error occurs
     */
    protected void doABPost(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException
    {
        // If this is not overridden, we invoke the raw HttpServlet "doGet" to
        // indicate that POST is not supported by this servlet.
        super.doPost(request, response);
    }



	/**
	 * This method handles exceptions
	 * 
	 * @param request	The users request
	 * @param response	The server response
	 * @param abrequest	The aidbase version of the request
	 * @param cause		The cause of the exception
	 * @param ex		The exception
	 * @throws ServletException If it has problems displaying the jsp page
	 * @throws IOException	If it has problems opening the jsp page
	 */
	public static void handleException(ServletRequest request,
	    HttpServletResponse response, Throwable exception) 
	throws ServletException, IOException 
	{
		if (exception instanceof GoDirectException) 
		{
	    	GoDirectException gde = (GoDirectException)exception;
	   		response.sendRedirect(gde.getURL());
	   		return;
	    }
		
		request.setAttribute("exception", exception);
		response.setStatus(500);
		String errorPage = null;
		
		{
			String message = null;
	
			if (exception instanceof RecursionDepthException) 
			{
				message = "something within the document has caused aidbase " +
					"to enter a loop.";
			} 
			else if (exception instanceof RecursionNestingException) 
			{
				errorPage = "/error/internal.jsp";
				message = "an internal error has occurred trying to filter " +
					"the specified document.";
			}
		
			if (message != null) {
				JSPManager.showInternalError(request, response, exception, message);
				return;
			}
		}
		
		if (exception instanceof RequestException || 
				exception instanceof DocumentNotFoundException) {
	        // A malformed URL exception is most likely caused
	        // by the user typing in a non-url into the url box.
	        // Respond nicely by showing the malformedurl jsp page
	        
			request.setAttribute("badurl",
					request.getParameter(WebUtils.getAidbaseUrlParam())); 

			errorPage = "/error/malformedurl.jsp";    
	    } 
		else if(exception instanceof ServerProtocolException) 
	    {
			
			String aburl = request.getAttribute("original.url.full").toString();
			
			if (aburl != null && (aburl.equals("") || aburl.equals("http://"))) {
				// user error: send them to bed with no supper
				errorPage = "/main/";
			}
			else
			{
				
				errorPage = "/error/filenotfound.jsp";
			}
		
	    }
		else if(exception instanceof ServerTimeoutException)
		{
			errorPage = "/error/timeout.jsp";
		}
		else if (exception instanceof BlockedUrlException) 
		{
	    	BlockedUrlException bue = (BlockedUrlException)exception;
	        URL rurl = bue.getRedirectUrl();
	        String rurlStr = null;
	        if (rurl != null) rurlStr = rurl.toString();
	        request.setAttribute("redirectUrl", rurlStr);
	        errorPage = "/error/blockurl.jsp";
		}
		else if (exception instanceof BadOutputException) 
		{
			BadOutputException boe = (BadOutputException)exception;
			String alturl = boe.getAlternateURL();
			String message = boe.getDisplayableMessage();
			request.setAttribute("alturl",alturl);
			request.setAttribute("message",message);
			errorPage = "/error/unabletofilter.jsp";
	    } 
		else if (exception instanceof IncompatibleDocumentException) 
		{
			errorPage = "/error/incompatibledocument.jsp"; 
	    }
		else
		{
			// Check first that the url entered was neither "" nor "http://"
			String aburl = request.getAttribute("original.url.full").toString();
			if (aburl != null && (aburl.equals("") || aburl.equals("http://"))) {
				// user error: send them to bed with no supper
				errorPage = "/main/";
			}
			else
			{
				// Cause unrecognised or not fixable by the user
				errorPage = "/error/internal.jsp";
			}
			
	    }
	
		LOG.warn("Showing JSP page "+errorPage+" in response to error "+
				exception);
	
		JSPManager.showJSP(request, response, errorPage);
	}
}    


