/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */

package org.aidworld.aidbase.webui.simulator;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Date;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.apache.log4j.Logger;

/**
 * Servlet filter to simulate a requested bandwidth.
 */

public class ThrottleFilter implements Filter {

    /** URL parameter to specify the required simulated bandwidth */
    public static final String BANDWIDTH_PARAM = "_ab_bps";

    /**
     * URL parameter used to share bandwidth between page loads.
     * Page requests using the same key parameter share the specified bandwidth, requests
     * using different keys have seperate bandwidth limits. The actual key string used
     * can be arbitrary.
     */
    public static final String THROTTLER_KEY_URL_PARAM = "_ab_key";

    /** Log4j logger */
    private static Logger LOG = Logger.getLogger(ThrottleFilter.class);

    /** Key prefix to use to store throttlers in the session */
    private static final String THROTTLER_SESSION_KEY = "ThrottleFilter";

    /** Perform the filter */
    public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain)
   		throws IOException, ServletException
    {
        HttpServletRequest httpRequest = (HttpServletRequest) request;

        boolean shouldThrottle = true;

        // Get the throttle speed from the url parameter
        int bps = -1;
        String strBandwidth = httpRequest.getParameter(BANDWIDTH_PARAM);
        if (strBandwidth == null) {
            shouldThrottle = false;
        } else {
            try {
                bps = Integer.parseInt(strBandwidth);
            } catch(NumberFormatException e) {
                if (LOG.isDebugEnabled()) LOG.debug("Bandwidth parameter invalid - "+e);
                shouldThrottle = false;
            }
        }
        if(bps < 300)
        {
        	shouldThrottle=false;
        }
       
        // Only throttle at all if a valid throttle bandwidth was supplied
        if (!shouldThrottle) {

            if (LOG.isDebugEnabled()) LOG.debug("Not throttling.");
            chain.doFilter(request, response);

        } else {

            String instanceKey = httpRequest.getParameter(THROTTLER_KEY_URL_PARAM);
            if (instanceKey == null) instanceKey = "";

            // Get filter instance from the session.
            HttpSession session = httpRequest.getSession(true);

            // If there isn't a filter in the session with this key, create one.
            String throttlerInstanceKey = THROTTLER_SESSION_KEY + instanceKey;
            Throttler throttler = (Throttler)session.getAttribute(throttlerInstanceKey);
            
            if (throttler == null || throttler.getBandwidthBps() != bps)
            {
                throttler = new Throttler(bps, throttlerInstanceKey);
                session.setAttribute(throttlerInstanceKey, throttler);
                if (LOG.isDebugEnabled()) LOG.debug(
                		"New Throttler instance created: key="
                		+ throttlerInstanceKey + ", bandwidth=" + bps);
            } 
            else
            {
            	if (LOG.isDebugEnabled()) LOG.debug(
            			"Using Throttler instance: key=" 
            			+ throttlerInstanceKey + ", from session=" 
						+ httpRequest.getSession() + ", shared bandwidth="
						+ throttler.getBandwidthBps() + " bps");
            }
            
            // Throttle the requested page with this throttler instance.
            LOG.info("request starting:" + httpRequest.getParameter("_ab_url"));
            long t=System.currentTimeMillis();
            throttler.throttle(request, response, chain);
            LOG.info("requeststopping:"+ (System.currentTimeMillis() - t) +
            		" "+ httpRequest.getParameter("_ab_url"));
        }
    }

    /** Need to provide this method to implement Filter */
    public void init(FilterConfig config) throws ServletException {
        if (LOG.isDebugEnabled()) LOG.debug("init");
    }

    /** Need to provide this method to implement Filter */
    public void destroy() {}

    /**
     * Inner class to perform the throttling. One Throttler instance can be shared by client requests
     * to limit the bandwidth, but more than one Throttler instance can be used by the same client
     * (e.g. to give two bandwidths windows to compare load times)
     */
    public static class Throttler{

        private static final int CHUNK_TO_WRITE = 512;
        private int m_bps;
        private String m_key;

        /** Create new Throttler instance to allow the specified bandwidth. */
        public Throttler(int bps, String key){
            m_bps = bps;
            m_key = key;
        }

        public int getBandwidthBps(){
            return m_bps;
        }

        /**
         * This method restricts the bandwidth. It is synchronized to make sure that multiple
         * requests throttled by the same throttler instance can't download in parallel.
         */
        private synchronized void throttle(ServletRequest request, ServletResponse response,
			FilterChain chain) 
            throws IOException, ServletException
        {
            if (LOG.isDebugEnabled()) LOG.debug("throttle: throttling");

            // example doFilter body stolen from HitCounterFilter example in
            // Sun Java docs [http://java.sun.com/products/servlet/Filters.html]

            Date startTime = new Date();

            CharResponseWrapper wrapper =
                new CharResponseWrapper( (HttpServletResponse)response );

            chain.doFilter(request, wrapper);

            byte[] responseContent = wrapper.getData();
            OutputStream out = response.getOutputStream();

            writeResponse(out, responseContent, startTime);

            out.flush();
        }

        public void writeResponse(OutputStream out, byte[] data, Date startTime) throws IOException {

            int contentLength = data.length;
            long totalBytesWritten = 0;

			if (LOG.isDebugEnabled())
				LOG.debug("writeResponse: Starting ("+m_key+
					") - size="+contentLength+" bytes");

            int index = 0;
            boolean writingComplete = false;
            while ( !writingComplete ) {

                int endIndex;
                if ( (index + CHUNK_TO_WRITE) > contentLength ) {
                    endIndex = contentLength;
                    writingComplete = true;
                } else {
                    endIndex = index + CHUNK_TO_WRITE;
                }

                out.write(data, index, (endIndex - index));

                totalBytesWritten += (endIndex - index);
                index = endIndex;

                sleep(totalBytesWritten, startTime);
                out.flush();
            }

            if (LOG.isDebugEnabled()){
                long bytes = totalBytesWritten;
                long timeTakenMillies = (new Date().getTime()) - (startTime.getTime());
                float speedbps = (bytes * 8000)/(timeTakenMillies+1);
                LOG.debug("writeResponse: Complete ("+m_key+
						") - size="+totalBytesWritten+" bytes; "+
						"time="+timeTakenMillies+"ms; "+
						"effective speed="+speedbps+" bps");
            }
        }

        private void sleep(long bytesWritten, Date startTime) {

            long actual_elapsed_ms =
                (new Date().getTime()) - (startTime.getTime());

            long expected_elapsed_ms =
                (bytesWritten * 1000 * 8) / m_bps;

            if (actual_elapsed_ms > expected_elapsed_ms){
                /* we are late, no need to throttle */
                //if (LOG.isDebugEnabled()) LOG.debug("sleep: we are late, returning. actual_elapsed_ms="+actual_elapsed_ms+" expected_elapsed_ms="+expected_elapsed_ms);
                return;
            }

            long sleep_time_ms = expected_elapsed_ms - actual_elapsed_ms;

            if (sleep_time_ms > 0) {
                //if (LOG.isDebugEnabled()) LOG.debug("sleep: sleeping for sleep_time_ms="+sleep_time_ms);
                try {
                    Thread.sleep( sleep_time_ms );
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }
    }
}
