/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */

package org.aidworld.aidbase.webui.util;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLEncoder;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.aidworld.aidbase.core.Version;
import org.aidworld.aidbase.util.Utils;
import org.apache.log4j.Logger;

/**
 * WebUI Utils class to perform certain utility functions. Used by both the
 * application and also directory by the JSP pages.
 */

public class WebUtils {

	private static Logger log = Logger.getLogger(WebUtils.class);

	private static String contextPath;

	static {
		try {
			contextPath = Utils.getAppProps()
					.getProperty("application.context");
		} catch (Exception e) {
			log.error(e);
		}
	}

	/**
	 * Returns the context path of the application, as configured in
	 * application.properties, for example "/loband"
	 * 
	 * @return The context path
	 */
	public static String getContextPath() {
		return contextPath;
	}

	/**
	 * Constructs the URL needed to send a link back through the loband HTML
	 * filter. You MUST prepend the context path to this URL, either with
	 * getContextPath() or getApplicationUrl[SSL](). This version returns the
	 * old 'page' which did not use to work properly with cookies
	 * 
	 * @return An absolute path MINUS the context, e.g. "/page"
	 * @todo this is no longer used? should be deleted
	 */
	public static String getFilterServletPath() {
		return "/page";
	}

	/**
	 * Constructs the URL needed to send a link back through the loband HTML
	 * filter. You MUST prepend the context path to this URL, either with
	 * getContextPath() or getApplicationUrl[SSL]().
	 * 
	 * @return An absolute path MINUS the context, e.g. "/filter"
	 * @todo this is no longer used? should be deleted
	 */
	public static String getRealFilterServletPath() {
		return "/filter";
	}

	// return a string formatted with 2 decimal places
	// nb doesn't do rounding.
	public static String formatDouble2dp(double x) {
		x = x * 100.0;
		int intpart = (int) x;
		String s2 = Integer.toString(intpart);
		String s = s2.substring(0, s2.length() - 2) + "."
				+ s2.substring(s2.length() - 2);

		// get rid of trailing zeros
		while (s.endsWith("0")) {
			s = s.substring(0, s.length() - 1);
		}
		if (s.endsWith(".")) {
			s = s.substring(0, s.length() - 1);
		}
		return s;
	}

	/**
	 * 
	 * @return The value of the getAidbasePipelineParam() for a binary pipeline,
	 *         e.g. "binary"
	 * @todo this is no longer used? should be deleted
	 */
	public static String getAidbaseBinaryType() {
		return "binary";
	}

	/**
	 * 
	 * @return The value of the getAidbasePipelineParam() for a script pipeline,
	 *         e.g. "script"
	 * @todo this is no longer used? should be deleted
	 */
	public static String getAidbaseScriptType() {
		return "script";
	}

	/**
	 * Get the name of the parameter used to specify a page request to the
	 * aidbase system.
	 * 
	 * @return The aidbase url parameter name, e.g. "_ab_url"
	 * @todo this is no longer used? should be deleted
	 */
	public static String getAidbaseUrlParam() {
		return getAidbaseParamName() + "url";
	}

	/**
	 * Get _ab_file - param to indicate file mode
	 * 
	 * @return _ab_file
	 */
	public static String getAidbaseFileParam() {
		return getAidbaseParamName() + "file";
	}

	/**
	 * Get the name of the parameter used to specify a page request to the
	 * aidbase system.
	 * 
	 * @return The aidbase port parameter name, e.g. "_ab_port"
	 */
	public static String getAidbasePortParam() {
		return getAidbaseParamName() + "port";
	}

	/**
	 * Get the name of the parameter used to tell Loband whether to use HTTP or HTTPS
	 * to retrieve the page.
	 * 
	 * @return The name of the security parameter, e.g. "_ab_secure"
	 */
	public static String getAidbaseSecurityParam() {
		return getAidbaseParamName() + "secure";
	}

	/**
	 * Get the name of the parameter used to tell Loband what kind of filtering to do
	 * on the document (e.g. HTML, binary, javascript)
	 * 
	 * @return The name of the filter type parameter, e.g. "_ab_type"
	 */
	public static String getAidbasePipelineParam() {
		return getAidbaseParamName() + "type";
	}

	/**
	 * Get the prefix for aidbase specific parameters.
	 * Parameters that start with this prefix are NOT sent
	 * to the remote website.
	 * 
	 * @return The prefix, e.g. "_ab_"
	 */
	public static String getAidbaseParamName() {
		return "_ab_";
	}

	/**
	 * Returns the base url host of a document url. This is used to work out the
	 * absolute urls of relative documents.
	 * 
	 * e.g. for the url 'http://www.google.com/test/test.html' this method will
	 * return 'http://www.google.com'
	 * 
	 * @param url
	 *            the absolute url of the document the url should be relative to
	 * @return the base url of the host
	 */
	public static String getBaseUrlHost(URL url) {
		String baseUrl;
		int port = url.getPort();
		if (port == -1) {
			baseUrl = url.getHost();
		} else {
			baseUrl = url.getHost() + ":" + port;
		}
		baseUrl = url.getProtocol() + "://" + baseUrl;
		return baseUrl;
	}

	/**
	 * Returns the path part of a relative url. This is used to work out the
	 * absolute urls of relative documents. This is included in the urls for
	 * relative documents, but not for slash-led relative urls.
	 * 
	 * e.g. for the url 'http://www.google.com/test/test.html' this method will
	 * return '/test'
	 * 
	 * @param url
	 *            the absolute url of the document the url should be relative to
	 * @return the base url of the host
	 */
	public static String getBaseUrlPath(URL url) {
		return url.getPath().replaceFirst("/[^/]*$", "");
	}

	/**
	 * Returns the host of the application.
	 * 
	 * @return The host of the application, such as "localhost"
	 * @throws Exception
	 *             If the properties file is wrongly configured
	 */
	public static String getApplicationHost() throws IOException {
		Properties applicationProperties = Utils.getAppProps();
		return (applicationProperties.getProperty("application.host",
				"localhost"));
	}

	/**
	 * Returns the URL of the application.
	 * 
	 * @return A complete url to the application, e.g
	 *         "http://localhost:8090/loband"
	 * @throws Exception
	 *             If the properties file is wrongly configured
	 */
	public static String getApplicationUrl() throws IOException {
		Properties applicationProperties = Utils.getAppProps();
		String applicationUrl = "http://"
				+ applicationProperties.getProperty("application.host",
						"localhost")
				+ ":"
				+ getApplicationPort()
				+ applicationProperties.getProperty("application.context",
						"/loband");
		return (applicationUrl);
	}

	public static String getApplicationPort() throws IOException {
	    String portNumber = System.getProperty("application.port");
	    if (portNumber == null)
	    {
	        Properties applicationProperties = Utils.getAppProps();
	        portNumber = applicationProperties.getProperty("application.port", "8080");
	    }
	    return portNumber;
	}

	/**
	 * Returns the URL of the application with the https protocol
	 * 
	 * @return A complete url to the application, e.g
	 *         "https://localhost:8090/loband"
	 * @throws Exception
	 *             If the properties file is wrongly configured
	 */
	public static String getApplicationUrlSSL() throws IOException {
		Properties applicationProperties = Utils.getAppProps();
		return ("https://"
				+ applicationProperties.getProperty("application.host",
						"localhost")
				+ ":"
				+ applicationProperties.getProperty("application.sslport",
						"8443") + applicationProperties.getProperty(
				"application.context", "/loband"));
	}

	/**
	 * Returns the aidbase standard page footer as html.
	 * 
	 * @param originalUrlFull
	 *            The url of the page that has been filtered
	 * @return The page footer
	 * @todo consolidate 3 versions of this function
	 */
	// Ideally, the method String getPageFooter(String originalUrlFull, boolean
	// includeWarning)
	// should be preferred to this method
	public static String getPageFooter(String originalUrlFull) {
		String footer = "";
		try {
			footer = "<table class=\"aidbar\"><tr>\n"
					+ "<td class=\"ableft\">"
					+ "&gt;(<a href=\""
					+ getApplicationUrl()
					+ "\"  title=\"Version: " // I18N 3011
					+ Version.getVersion()
					+ "\">"
					+ Version.getName()
					+ "</a>)-"
					+ "</td>\n"
					+ "<td class=\"abright\">"
					+ "designed by " // I18N 3012
					+ "(<a href=\"http://www.aidworld.org\">aidworld</a>)</td>\n"
					+ "</tr></table>\n";
		} catch (Exception e) {

			e.printStackTrace();
		}
		return footer;
	}

	/**
	 * Returns the aidbase standard page footer as html.
	 * 
	 * @return The page footer
	 * @todo consolidate 3 versions of this function
	 */
	public static String getPageFooter(boolean includeWarning) {
		String footer = "";
		try {
			footer = "<table class=\"aidbar\"><tr>\n" + "<td class=\"ableft\">"
					+ "&gt;(<a href=\"" + getContextPath()
					+ "\" target=\"_top\" " + "title=\"Version: "
					+ Version.getVersion() // I18N 3011
					+ "\">" + Version.getName() + "</a>)-" + "</td>\n";
			if (includeWarning) {
				footer += "<td class=\"abmid\">"
						+ "This page might not display properly." + "</td>\n";
			}
			footer += "<td class=\"abright\">" + "designed by " // I18N 3012
					+ "(<a target=\"_top\" "
					+ "href=\"http://www.aidworld.org\">aidworld</a>)</td>\n"
					+ "</tr></table>\n" + "</body>\n" + "</html>";
		} catch (Exception e) {

			e.printStackTrace();
		}
		return footer;
	}

	public static String getPageFooter() {
		return getPageFooter (false);
	}

	/**
	 * Returns the page header as HTML.
	 * 
	 * @param originalUrlFull
	 *            The url of the page that has been filtered
	 * @return The page header
	 * @todo consolidate 3 versions of this function
	 */
	public static String getPageHeader(String originalUrlFull) {
		String originalUrlShort = originalUrlFull;

		if (originalUrlFull != null) {
			if (originalUrlFull.indexOf("_ab_locinfo") > -1) {
				originalUrlFull = originalUrlFull.substring(originalUrlFull
						.indexOf("_ab_locinfo") + 12);
				originalUrlShort = originalUrlFull;
			}

			if (originalUrlShort.length() > 30) {
				originalUrlShort = originalUrlShort.substring(0, 30) + "...";
			}
		}

		try {
			return "\n"
					+ "<form action=\""
					+ getContextPath()
					+ getFilterServletPath()
					+ "\" method=\"get\" "
					+ "target=\"_top\">\n"
					+ "<table class=\"aidbar\">\n"
					+ "<tr>\n"
					+ "<td class=\"ableft\">"
					+ "&gt;(<a href=\""
					+ getContextPath()
					+ "\" target=\"_top\" "
					+ "title=\"Version: " // I18N 3011
					+ Version.getVersion()
					+ "\">"
					+ Version.getName()
					+ "</a>)-"
					+ "</td>\n"
					+ "<td class=\"abmid\">\n"
					+ "<input class=\"aburl\" type=\"text\""
					+ " name=\""
					+ getAidbaseUrlParam()
					+ "\""
					+ " value=\""
					+ (originalUrlFull == null ? "http://" : originalUrlFull)
					+ "\">\n"
					+ "<input type=\"submit\" name=\""
					+ getAidbaseParamName()
					+ "request\" value=\"Go\">\n"
					+ "</td>\n"
					+ "<td class=\"abright\">"
					+ (originalUrlFull == null ? "&nbsp;" : "<a href=\""
							+ originalUrlFull + "\" target=\"_top\">"
							+ "original</a> |\n" // I18N 3013
							+ "<a href=\"" + getContextPath()
							+ "/badpagereport?_page="
							+ URLEncoder.encode(originalUrlFull, "UTF-8")
							+ "\"" + " target=\"_new\">Report error</a>\n") // I18N
																			// 3014
					+ "</td>\n" + "</tr>\n" + "</table>\n" + "</form>\n";
		} catch (Exception e) {
			throw new Error("Unsupported encoding: UTF-8", e);
		}
	}

	// @todo consolidate 3 versions of this function
	
	public static String getPageHeaderSimulator(String originalUrlFull,
			Properties params) {
		String originalUrlShort = originalUrlFull;
		String blurb = "";
		if (originalUrlFull != null) {
			if (originalUrlFull.indexOf("_ab_locinfo") > -1) {
				originalUrlFull = originalUrlFull.substring(originalUrlFull
						.indexOf("_ab_locinfo") + 12);
				originalUrlShort = originalUrlFull;
			}

			if (originalUrlShort.length() > 30) {
				originalUrlShort = originalUrlShort.substring(0, 30) + "...";
			}
		}

		if (params.get("aidbase.filter.disable") != null) {
			blurb = "Unfiltered Page. ";
		} else {
			blurb = "Filtered Page. ";
		}

		String x = params.get("aidbase.simulatorOnOff").toString();
		if (x.equals("true")) {
			double kbps;
			String bps = (String) (params.get("aidbase.simulatorbps"));
			try {
				kbps = Double.parseDouble(bps) / 1000.0;
			} catch (Exception e) {
				kbps = 0.0;
			}

			blurb = blurb + "Simulated Bandwidth: " + formatDouble2dp(kbps)
					+ "Kbps";
		}

		try {
			return "\n" + "<table class=\"aidbar\">\n"
					+ "<tr class=\"ableft\">\n" + "<td class=\"ableft\">"
					+ "&gt;(<a href=\"" + getApplicationUrl()
					+ "\" target=\"_top\" " + "title=\"Version: " // I18N 3011
					+ Version.getVersion() + "\">" + Version.getName()
					+ "</a>)-" + "</td>\n" + "<td class=\"abmid\">\n" + blurb
					+ "</td><td class=\"abright\">" + "\n" // I18N 3013
					+ "<a href=\"" + getApplicationUrl()
					+ "badpagereport?_page="
					+ URLEncoder.encode(originalUrlFull, "UTF-8") + "\""
					+ " target=\"_new\">Report error</a>&nbsp;\n" // I18N 3014
					+ "</td>\n" + "</tr>\n" + "</table>\n";
		} catch (Exception e) {
			throw new Error("Unsupported encoding: UTF-8", e);
		}
	}

	/**
	 * @return the HTML to be inserted at end of page header
	 * @todo consolidate 3 versions of this function
	 */
	public static String getPageHeadHtml() {
		return "<meta http-equiv=\"content-type\" "
				+ "content=\"text/html; charset=UTF-8\" />\n"
				+ "<meta name=\"robots\" content=\"index,nofollow\">\n"
				+ "<link rel=\"stylesheet\" href=\"" + getContextPath()
				+ "/loband.css\">\n";
	}

	public static String getPageHeadHtmlNoFilter() {
		return "<meta http-equiv=\"content-type\" "
				+ "content=\"text/html; charset=UTF-8\" />\n"
				+ "<meta name=\"robots\" content=\"index,nofollow\">\n"
				+ "<link rel=\"stylesheet\" href=\"" + getContextPath()
				+ "/lobandnofilter.css\">\n";
	}

	/**
	 * Take an url and a method and return the url of the servlet to call this
	 * address with a pipeline of the method type, e.g. with originalUrl =
	 * "http://www.bbc.co.uk/bbc.gif" and method = "binary", the method returns
	 * something like
	 * "/loband/filter/bbc.co.uk/www/bbc.gif?_ab_type=binary"
	 * 
	 * @param originalUrl
	 * @param method
	 * @return the correct url to call the servlet reading originalUrl with a
	 *         pipeline of type method
	 * @throws IOException
	 */
	public static String getLobandCallUrl(String originalUrl, String method)
			throws IOException {
		String newUrl = "";
		
		Pattern pat3=Pattern.compile("http://",Pattern.CASE_INSENSITIVE);
		Matcher mat3=pat3.matcher(originalUrl);
		if (mat3.find()&&mat3.start()==0)
			originalUrl="http://"+originalUrl.substring(7);
		
		Pattern pat4=Pattern.compile("https://",Pattern.CASE_INSENSITIVE);
		Matcher mat4=pat4.matcher(originalUrl);
		if (mat4.find()&&mat4.start()==0)
			originalUrl="https://"+originalUrl.substring(8);
		
		boolean isSecure = false;
		if (originalUrl.startsWith("https://")) {
			isSecure = true;
			return originalUrl;
			//originalUrl = originalUrl.substring(7);
		} else if (originalUrl.startsWith("http://"))
			originalUrl = originalUrl.substring(6);
		if (!originalUrl.startsWith("/"))
			originalUrl = "/" + originalUrl;
		
		// start looking for the port if any
		Pattern pat = Pattern.compile("\\:(\\d+)");
		Matcher mat = pat.matcher(originalUrl);
		String sPort = "";
		int nQuestionIndex = originalUrl.indexOf("?", 1);

		if (mat.find()) {
			int nColonIndex = originalUrl.indexOf(":", 1);
			if (nColonIndex != -1) {
				int nSlashIndex = originalUrl.indexOf("/", 1);
				if (nSlashIndex > nColonIndex || nSlashIndex == -1) {
					if (nQuestionIndex > nColonIndex || nQuestionIndex == -1) {
						sPort = originalUrl.substring(mat.start(1), mat.end(1));
						originalUrl = originalUrl.substring(0, mat.start())
								+ originalUrl.substring(mat.end());
					}
				}
			}
		}
		if(originalUrl.startsWith("/"))
		{
			originalUrl = originalUrl.substring(1);
		}
		originalUrl = convertURLToLobandFormat(originalUrl);
		
		newUrl = getContextPath() + getRealFilterServletPath() + "/" + originalUrl;
		
		String append = "";
		
		if(!sPort.equals(""))
			append += getAidbasePortParam() + "=" + sPort + "&";
		if(!method.equals(""))
			append += getAidbasePipelineParam() + "=" + method;

		if (append.length() > 0)
		{
			if(nQuestionIndex == -1)
			{
				newUrl += "?" + append;
			} else {
				newUrl += "&" + append;
			}
		}
		
		return newUrl;
	}

	/**
	 * Takes a url and return the same url where the parameters have been
	 * encoded in UTF-8
	 * 
	 * @param url
	 * @return The same
	 */
	public static String endodeParam(String url)
			throws UnsupportedEncodingException {
		String encodedUrl = null;
		int position = url.indexOf('?');
		if (position == -1) {
			encodedUrl = url;
		} else {
			encodedUrl = url.substring(0, position + 1)
					+ URLEncoder.encode(url.substring(position + 1), "UTF-8");
		}
		return encodedUrl;
	}

	public static String getMainPageTestString() {

		return "Enter the address of the web page* to simplify in the box above.";

	}

	public static String getFileNotFoundPageTestString() {
		
		return "File Not Found";
	}

	/**
	 * Takes string of the form http://www.robin.com:80/hello
	 * and returns string of the form http://www.robin.com/hello&_ab_port=80
	 */
	public static String convertPortToParam(String aburl)
	{
		URL url;
		try
		{
			if(!aburl.startsWith("http://") && !aburl.startsWith("https://"))
				url = new URL("http://" + aburl);
			else
				url = new URL(aburl);
			int port = url.getPort();
			// remove the "http://" at the begining of the url
			if(aburl.startsWith("http://"))
				aburl = aburl.substring(7);
			else if(aburl.startsWith("https://"))
				aburl = aburl.substring(8);
			
			// remove the port from the url if it was there
			// and add it as an _ab_port= parameter
			if(port != -1)
			{
				aburl = aburl.replaceFirst(":\\d+", "");
				if(aburl.indexOf("?")!=-1)
				{
					aburl=aburl + "&_ab_port=" + Integer.toString(port);
				}
				else
				{
					aburl=aburl + "?_ab_port=" + Integer.toString(port);
				}
			}
		}
		catch (Exception e)
		{
			
		}
		
		return aburl;
	}
	

	/**
	 * Takes a url like "www.bbc.com/path" and returns "com/bbc/www/%20/path"
     * which will used in the request to FilterServlet. This is to ensure
     * that Cookies with domain bbc.co.uk are sent on both www.bbc.co.uk
     * and news.bbc.co.uk. If there is no '.' in the first part of the url, as in
     * "localhost/loband/test", then the method inserts "%20" between the two '/'
     * as in "localhost/%20/loband/test".
     * 
	 * @param sURL of the form www.bbc.com/path
	 * @return URL of the form bbc.com/www/path
	 */  
	public static String convertURLToLobandFormat(String sURL)
	{
		
		int firstSlash = sURL.indexOf("/"); 
		
		String domain;
		
		String path;
		
		if (firstSlash != -1)
		{
			
			domain = sURL.substring(0, firstSlash);
			
			path = sURL.substring(firstSlash);
			
		}
		else
		{
			
			domain = sURL;
			
			path = "";
			
		}
				
		String[] nuggets = domain.split("\\.");
		
		StringBuffer newURL = new StringBuffer("");
		
		for(int i = nuggets.length - 1; i>=0 ; i--)
		{
			
			newURL.append(nuggets[i]);
			if(i>0)
			{
				newURL.append("/");
			}
			
		}
		
		if(!path.equals(""))
		{
			newURL.append("/%20");
			newURL.append(path);
		}
		
		
		return newURL.toString();
		
	}

    /**
     * Undo what was done in convertURLToLobandFormat(String sURL). The string sURL
     * should not start by a '/'
     * 
     * @param sURL of the form "uk/co/bbc/www/%20/news"
     * @return String of the form "www.bbc.co.uk/news"
     */
    public static String unConvertURLToLobandFormat(String sURL)
    {
    	
    	int pathSplitter = sURL.indexOf("/%20");
    	
    	String domain;
    	
    	String path;
    	
    	if(pathSplitter != -1)
    	{
    		
    		domain = sURL.substring(0, pathSplitter);
    		
    		path = sURL.substring(pathSplitter + 4);
    		
    	}
    	else
    	{
    		
    		domain = sURL;
    		
    		path = "";
    		
    	}
    	
    	String[] nuggets = domain.split("/");
    	
    	StringBuffer newURL = new StringBuffer("");
    	
    	for(int i = nuggets.length-1; i>=0; i--)
    	{
    		
    		newURL.append(nuggets[i]);
    		if(i>0)
    		{
    			
    			newURL.append(".");
    			
    		}
    		
    	}
    	
    	newURL.append(path);
    	
    	return newURL.toString();
    	
    }
}