/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */

package org.aidworld.aidbase.core;

import java.net.URL;
import java.net.URLEncoder;
import java.util.Hashtable;

import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.aidworld.aidbase.document.Document;
import org.aidworld.aidbase.document.HtmlDocument;
import org.aidworld.aidbase.factory.StandardDocumentFactory;
import org.aidworld.aidbase.filter.SimplifyPipeline;
import org.aidworld.aidbase.request.Request;
import org.aidworld.aidbase.request.TestRequest;
import org.apache.log4j.Logger;

import com.meterware.httpunit.GetMethodWebRequest;
import com.meterware.httpunit.WebConversation;
import com.meterware.httpunit.WebRequest;
import com.meterware.httpunit.WebResponse;

/**
 * This isn't really a JUnit test. You can create a Spec
 * with a given URL, specifying what should be in the document,
 * how large it should be, etc. and then verify that
 * after processing it really does conform with the spec.
 * @todo rewrite as a real unit test.
 */

public class DocumentSpec extends TestCase {
	private String urlString = null;
	private String[] expectedPresent = null;
	private String[] expectedAbsent = null;
	//      size of <html><head></head><body></body></html>
	private int minimumSize = 39;
	private int maximumSize = 204800; // 200 KB
	// private float maximumTime = 30; // 30 seconds max by default

    private Logger LOG = Logger.getLogger(DocumentSpec.class);
    
	public DocumentSpec (
	        String name,
			String urlString,
			String[] expectedPresent,
			String[] expectedAbsent) {

	    // this.setName(name);
		this.urlString = urlString;
		this.expectedPresent = expectedPresent;
		this.expectedAbsent = expectedAbsent;
	}

	public DocumentSpec (
	        String name,
			String urlString,
			String[] expectedPresent,
			String[] expectedAbsent,
			int minimumSize,
			int maximumSize) {
	    this(name, urlString, expectedPresent, expectedAbsent);
		this.minimumSize = minimumSize;
		this.maximumSize = maximumSize;
	}

	/** The URL of the original document, as passed to the constructor.
	 * @return a URL as a string
	 */
	public String getUrlString() {
		return urlString;
	}

	/** An array of strings that are expected in the filtered output,
	 * as passed to the constructor.
	 * @return an array of strings
	 */
	public String[] getExpectedPresent() {
		return expectedPresent;
	}

	/** An array of string that are NOT expected in the filtered output,
	 * as passed to the constructor.
	 * @return an array of strings
	 */
	public String[] getExpectedAbsent() {
		return expectedAbsent;
	}

	/** Returns the smallest sensible size of the filtered document,
	 * as passed to the constructor.
	 * @return size in bytes
	 */
	public int getMinimumSize() {
		return minimumSize;
	}

	/** Returns the largest sensible size of the filtered document,
	 * as passed to the constructor.
	 * @return size in bytes
	 */
	public int getMaximumSize() {
		return maximumSize;
	}
	
	/** Returns the maximum number of seconds that the request can take to process
	 * before the test will fail with an error (this only happens after the request is finished)
	 * @return the current value in seconds
	 */
	/*
	public float getMaximumTime() {
		return maximumTime;
	}
	*/

	/** Sets the maximum number of seconds that the request can take to process
	 * before the test will fail with an error (this only happens after the request is finished)
	 * @return nothing
	 */
	/*
	public void setMaximumTime(float maxTime) {
		this.maximumTime = maxTime;
	}
	*/
	
	private Document filteredDoc;
	
    public void setUp() throws Exception {
        URL url = new URL(this.urlString);
        if (httpUnitUrl !=null){
        	WebConversation conversation = new WebConversation();
        	WebRequest request = new GetMethodWebRequest(httpUnitUrl+
        			"&_ab_url="+URLEncoder.encode(this.urlString,"UTF-8"));
        	WebResponse response = conversation.getResponse( request );
        	this.filteredDoc = new HtmlDocument(response.getText(),new Hashtable());
        	LOG.debug(response.getText());
        }else{
	        StandardDocumentFactory sdf = new StandardDocumentFactory();
	        SimplifyPipeline pipeline = new SimplifyPipeline(sdf);
	        Request request = new TestRequest(url, "http://www.aidbase.org", null);
	        
	        try {
	        	this.filteredDoc = pipeline.service(request);
	        } catch (Throwable e) {
	        	throw new Exception("Test failed for "+url, e);
	        }
        }
    }

	/** Verifies that the size of the document is greater than the minimum size.
	 * 
	 * @param document the document to test
	 * @return whether the test passed or failed
	 */
	public void verifyShouldNotBeLessThanMinimumSize() {
	    String docString = filteredDoc.toString();
	    assertNotNull("Document is null: "+urlString, docString);

	    assertTrue(
	            "Document too small, size (" 
	            + docString.length() 
	            + " bytes) is less than minimum expected size ("
	            + minimumSize + " bytes) for "+urlString,
	            docString.length() >= minimumSize);
	}

	/** Verifies that the size of the document is less than the maximum size.
	 * 
	 * @param document the document to test
	 * @return whether the test passed or failed
	 */
	public void verifyShouldNotBeGreaterThanMaximumSize() {
	    String docString = filteredDoc.toString();
	    assertNotNull("Document is null: "+urlString, docString);

	    assertTrue(
	            "Document too large, size (" 
	            + docString.length() 
	            + " bytes) is greater than maximum expected size ("
	            + maximumSize + " bytes) for "+urlString,
	            docString.length() <= maximumSize);
	}

	/** Verifies the presence of all the expected strings within the document.
	 * 
	 * @param document the document to test
	 * @return whether the test passed or failed
	 */
	public void verifyExpectedPresent() {
		if (expectedPresent == null || expectedPresent.length == 0) {
			return;
		}
		
		String docString = filteredDoc.toString();
		assertNotNull("Document is null: "+urlString, docString);

		for (int i = 0; i < expectedPresent.length; i++) {
			assertTrue(
					"Expected string '"
					+ expectedPresent[i]
					+ "' not found in document "+urlString,
					docString.indexOf(expectedPresent[i]) != -1);
		}
	}


	/** Verifies the absence of all the unexpected strings within the document.
	 * 
	 * @param document the document to test
	 * @return whether the test passed or failed
	 */
	public void verifyExpectedAbsent() {
		if (expectedAbsent == null || expectedAbsent.length == 0) {
			return;
		}
		
		String docString = filteredDoc.toString();
		assertNotNull("Document is null: "+urlString, docString);

		for (int i = 0; i < expectedAbsent.length; i++) {
			assertTrue(
					"Unexpected string '"
					+ expectedAbsent[i]
					+ "' found in document "+urlString,
					docString.indexOf(expectedAbsent[i]) == -1);
		}
	}
	
	public void testDocumentShouldConformToSpec() throws Exception {
	    verifyShouldNotBeLessThanMinimumSize();
	    verifyShouldNotBeGreaterThanMaximumSize();
	    verifyExpectedPresent();
	    verifyExpectedAbsent();
	}
	
	public TestSuite getSuite() throws Exception {
	    TestSuite suite = new TestSuite("anonymous");
	    
	    DocumentSpec clone = new DocumentSpec(
	            this.getName(),
	            this.getUrlString(),
	            this.getExpectedPresent(),
	            this.getExpectedAbsent(),
	            this.getMinimumSize(),
	            this.getMaximumSize());
	    clone.setName("testDocumentShouldConformToSpec");
	    clone.setHttpUnitUrl(this.httpUnitUrl);
	    suite.addTest(clone);
	    
	    return suite;
	}
	
	String httpUnitUrl = null;
	
	public void setHttpUnitUrl(String url){
		this.httpUnitUrl = url;
	}
}
