/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */

package org.aidworld.aidbase.offline.unit;

import java.io.FileNotFoundException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Hashtable;

import junit.framework.Test;
import junit.framework.TestSuite;

import org.aidworld.aidbase.core.BasedTest;
import org.aidworld.aidbase.document.Document;
import org.aidworld.aidbase.document.HtmlDocument;
import org.aidworld.aidbase.error.AidbaseException;
import org.aidworld.aidbase.factory.StandardDocumentFactory;
import org.aidworld.aidbase.request.Request;
import org.aidworld.aidbase.request.TestRequest;
import org.aidworld.aidbase.webui.util.WebUtils;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.methods.GetMethod;

/**
 * Unit tests for the StandardDocumentFactory class.
 */

public class TestStandardDocumentFactory extends BasedTest {
	public TestStandardDocumentFactory() {
	    super();
	}

	public TestStandardDocumentFactory( String name ) {
	    super( name );
	}

	public static void main(String args[]) {
		junit.textui.TestRunner.run( suite() );
	}
	
	public static Test suite() {
  	    return new TestSuite(TestStandardDocumentFactory.class);
	}
	
	public void testShouldExtractContentTypeFromHeader() throws Exception {
		String surl = WebUtils.getApplicationUrl() + 
			"/test/encoding/content-type-header.jsp";
		URL url = new URL(surl);
		Request request = new TestRequest(url, "/foo", new Hashtable());
		StandardDocumentFactory factory = new StandardDocumentFactory();
		HtmlDocument doc = (HtmlDocument)( factory.service(request) );
		assertEquals("windows-1251", doc.getOriginalEncoding());
	}
	
	public void testShouldReturnNullWithoutContentTypeHeader() 
	{
		String [] testPatterns = new String[]{
				"text/plain; charset=",
				"text/plain; charset",
				"text/plain; ",
				"text/plain",
		};
		
		for (int i = 0; i < testPatterns.length; i++) {
			String type = testPatterns[i];
			
			String encoding = 
				StandardDocumentFactory.getEncodingFromContentType(type);

			assertEquals(null, encoding); 
		}
	}

	public void testShouldExtractCorrectEncodingFromContentType() 
	{
		String [] testPatterns = new String[]{
				"text/plain;charset=FOO-8",
				"text/plain; charset=FOO-8",
				"text/plain ; charset = FOO-8 ",
				"text/plain; foo=bar; charset=FOO-8",
				"text/plain; foo=bar; charset=FOO-8; foo=bar",
				"text/plain; foo=bar;;charset=FOO-8;;foo=bar",
		};
		
		for (int i = 0; i < testPatterns.length; i++) {
			String type = testPatterns[i];
			
			String encoding = 
				StandardDocumentFactory.getEncodingFromContentType(type);

			assertTrue("Failed to match encoding in <"+type+">",
					encoding != null);
			assertEquals("Wrong content-type in <"+type+">",
					"FOO-8", encoding); 
		}
	}

	public void testShouldExtractCorrectContentTypeFromBody() 
	{
		String [] testPatterns = new String[]{
				"<meta http-equiv=\"content-type\" content=\"text/plain; charset=FOO-8\">",
				"<meta http-equiv = \"content-type\" content = \"text/plain; charset = FOO-8 \">",
				"<meta http-equiv = content-type content = text/plain; charset = FOO-8 >",
				"<meta http-equiv=content-type content=text/plain; charset=FOO-8 >",
				"<meta http-equiv='content-type' content='text/plain; charset=FOO-8'>",
		};
		
		for (int i = 0; i < testPatterns.length; i++) {
			String tag = testPatterns[i];
			
			String type = 
				StandardDocumentFactory.getContentTypeFromBody(tag);
			
			assertTrue("Wrong content type '"+type+"' in '"+tag+"'", 
					type.startsWith("text/plain"));
			
			String encoding = 
				StandardDocumentFactory.getEncodingFromContentType(type);

			assertEquals("Extracted wrong content type from "+
					"<"+type+"> in "+tag, "FOO-8", encoding); 
		}
	}
	

	public void testTestUriSchemeShouldWork() throws URISyntaxException
	{
		URI uri = new URI("test:/foo/bar");
		assertTrue(uri.getPath().equals("/foo/bar"));
	}
	
	
	//Check to see if HttpClient works properly and allows
	//us to change the user-agent.
	public void testShouldBeAbleToChangeUserAgent() throws Exception
	{
		String url = getAidbaseUrl() + "/test/userAgent.jsp";
		HttpClient client = new HttpClient();
		HttpMethod method = new GetMethod(url);
		
		method.setRequestHeader("User-Agent", "FOOBAR");
		client.executeMethod(method);
		
		String response = method.getResponseBodyAsString();
		
		assertEquals("FOOBAR", response);
	} 
	
	public void testAidbaseGeneratesRandomUserAgentStrings()
	throws Exception 
	{
		response = getThroughAidbase(getAidbaseUrl() + "/test/userAgent.jsp");
		String agent1 = response.getText()
		.replaceAll("(?s).*</table>", "")
		.replaceAll("(?s)</body></html>.*", "");
		
		response = getThroughAidbase(getAidbaseUrl() + "/test/userAgent.jsp");
		String agent2 = response.getText()
		.replaceAll("(?s).*</table>", "")
		.replaceAll("(?s)</body></html>.*", "");
		
		if (agent1.indexOf("httpunit") == -1) {
			fail("user agent was not passed through (should be 'httpunit'): "
					+ agent1);
		}
		
		if (agent2.indexOf("httpunit") == -1) {
			fail("user agent was not passed through (should be 'httpunit'): "
					+ agent2);
		}
		
		if (agent1.equals(agent2)) {
			fail("returned same user agent: "+agent1);
		}
	}
	
	public void testRetrieveNonexistantDocumentShouldThrowException() 
	throws Exception 
	{
		AidbaseException ae = null;
	
		try {
			StandardDocumentFactory fac = new StandardDocumentFactory();
			TestRequest req = new TestRequest(
					new URL("file:/no-such-file.html"),
					"/aidbase", new Hashtable());
			fac.service(req);
		} catch (AidbaseException e) {
			ae = e;
		}
		
		assertNotNull("StandardSocumentFactory did not throw an exception", ae);
		
		Throwable cause = ae.getCause();
		assertNotNull("Exception does not have a cause", cause);
		assertTrue("Exception cause is wrong type ("+cause+")",
				cause instanceof FileNotFoundException);
		
		assertEquals("Wrong file name", "no-such-file.html",
				cause.getMessage());
	}


	public void testRetrieveTestPageShouldWork() throws Exception 
	{
		StandardDocumentFactory fac = new StandardDocumentFactory();
		TestRequest req = new TestRequest(
				new URL("file:/www.google.fr.html"),
				"/aidbase", new Hashtable());
		fac.service(req);
	}

	
	public void testCharacterSetConversionShouldWork() throws Exception 
	{
		byte [] correctBytes = new byte[]{
				(byte)0xc2,(byte)0xbc,(byte)0x20,(byte)0xc2,
				(byte)0xbd,(byte)0x20,(byte)0xc2,(byte)0xbe,
				(byte)0x20,(byte)0xc2,(byte)0xbf,(byte)0x20,
				(byte)0xc3,(byte)0x80,(byte)0x20,(byte)0xc3,
				(byte)0x81,(byte)0x20,(byte)0xc3,(byte)0x82,
				(byte)0x20,(byte)0xc3,(byte)0x83,(byte)0x20,
				(byte)0xc3,(byte)0x84,(byte)0x20,(byte)0xc3,
				(byte)0x85,(byte)0x20,(byte)0xc3,(byte)0x86,
				(byte)0x20,(byte)0xc3,(byte)0x87,(byte)0x20,
				(byte)0xc3,(byte)0x88,(byte)0x20,(byte)0xc3,
				(byte)0x89,(byte)0x20,(byte)0xc3,(byte)0x8a,
				(byte)0x20,(byte)0xc3,(byte)0x8b,(byte)0x20,
				(byte)0xc3,(byte)0x8c,(byte)0x20,(byte)0xc3,
				(byte)0x8d,(byte)0x20,(byte)0xc3,(byte)0x8e,
				(byte)0x20,(byte)0xc3,(byte)0x8f,(byte)0x20,
				(byte)0xc3,(byte)0x90,(byte)0x20,(byte)0xc3,
				(byte)0x91,(byte)0x20,(byte)0xc3,(byte)0x92,
				(byte)0x20,(byte)0xc3,(byte)0x93,(byte)0x20,
				(byte)0xc2,(byte)0xb1,(byte)0x20,(byte)0xc3,
				(byte)0xb1,(byte)0x0a,
		};
		
		StandardDocumentFactory fac = new StandardDocumentFactory();
		
		String testPageName = "encoding-iso-8859-1.html";
		URL testPage = fac.getFileUrl(testPageName);
		
		if (testPage == null) {
			// System.err.println(fac.getFileUrl("").toString());
			// System.err.println(fac.getFileUrl(testPageName).toString());
			throw new FileNotFoundException(
					fac.getFileUrl("").toString() + testPageName);
		}
		
		String correctOutput = "<html><head>\n"
			+ "<meta http-equiv=\"content-type\"" 
			+ " content=\"text/html; charset=ISO-8859-1\">\n"
			+ "</head><body>\n"
			+ new String(correctBytes, "UTF-8")
			+ "</body></html>\n";
		
		TestRequest req = new TestRequest(
				new URL("file:/" + testPageName),
				"/aidbase", new Hashtable());
		Document doc = fac.service(req);

		char[] chars= new String(correctBytes, "UTF-8").toCharArray();
		StringBuffer output = new StringBuffer();
		for(int i=0; i<chars.length;i++){
			output.append(Integer.toHexString((int)chars[i])+ " ");
		}
		System.out.println("A: "+output);
//		System.out.println("doc.toString(): "+doc.toString());
		System.out.println("B: "
				+ "c2 bc 20 c2 bd 20 c2 be 20 c2 bf 20 c3 80 20 c3 81 20 "
				+ "c3 82 20 c3 83 20 c3 84 20 c3 85 20 c3 86 20 c3 87 20 "
				+ "c3 88 20 c3 89 20 c3 8a 20 c3 8b 20 c3 8c 20 c3 8d 20 "
				+ "c3 8e 20 c3 8f 20 c3 90 20 c3 91 20 c3 92 20 c3 93 20 "
				+ "c2 b1 20 c3 b1 0a");
		chars= doc.toString().toCharArray();
		output = new StringBuffer();
		for(int i=0; i<chars.length;i++){
			output.append(Integer.toHexString((int)chars[i])+ " ");
		}
		// System.out.println("C: "+output);
				 
		assertEquals("Character set conversion failed", correctOutput, doc.toString());
		
		testPageName = "encoding-utf-8.html";
		testPage = fac.getFileUrl(testPageName);
		
		if (testPage == null) {
			// System.err.println(fac.getFileUrl("").toString());
			// System.err.println(fac.getFileUrl(testPageName).toString());
			throw new FileNotFoundException(
					fac.getFileUrl("").toString() + testPageName);
		}
		
		correctOutput = "<html><head>\n"
			+ "<meta http-equiv=\"content-type\"" 
			+ " content=\"text/html; charset=UTF-8\">\n"
			+ "</head><body>\n"
			+ new String(correctBytes, "UTF-8")
			+ "</body></html>\n";
		
		req = new TestRequest(
				new URL("file:/" + testPageName),
				"/aidbase", new Hashtable());
		doc = fac.service(req);

		assertEquals("Character set passthrough failed", correctOutput, doc.toString());
		
	}
}
