/* ------------------------------------------------------------------------
 * Aidbase simplifies HTML pages to reduce bandwidth usage, et cetera.
 * Copyright (C) 2002-2004 Aidworld Humanitarian ICT, Cambridge, UK. 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * This software is intended to be as accessible as possible. If acceptance 
 * of this license conflicts with your interests, please contact us to 
 * discuss how we might be able to accommodate your requirements. 
 * 
 * Please look at our website, http://www.aidworld.org, to find out more 
 * about our organisation and to read our mission statement. 
 * ------------------------------------------------------------------------
 */

package org.aidworld.aidbase.performance;

import java.util.Vector;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.aidworld.aidbase.core.DocumentSpec;
import org.aidworld.aidbase.offline.unit.OfflineDocumentSpecs;
import org.apache.log4j.Logger;

/**
 * TestConcurrency contains unit tests for reasonable
 * concurrent performance
 */

public class TestConcurrency extends TestCase {
    //These tests pass reference documents through the pipeline to 
    //check the output. They load local files.
    
    private Logger LOG = Logger.getLogger(TestConcurrency.class);

	class TestThread extends Thread {
		private DocumentSpec spec;
		private long timeTotal;
		private long timePerRun;
		private long numPasses;
		
		public TestThread(DocumentSpec spec) {
			this.spec = spec;
		}
		
	    public void run() {
	    	try {
		        this.numPasses = 10;
		        
				long startMillis = System.currentTimeMillis();      
   				for (int pass = 0; pass < numPasses; pass++) {
					spec.run();
				}
				long endMillis = System.currentTimeMillis();
				        
		        this.timeTotal  = endMillis - startMillis;
		        this.timePerRun = this.timeTotal / numPasses;
	    	} catch (Exception e) {
	    		// force test failure
	    		TestCase.assertTrue(e.toString(), false);
	    	}
	    }
	    
	    public long getTimeTotal() {
	    	return this.timeTotal;
	    }

	    public long getTimePerRun() {
	    	return this.timePerRun;
	    }

	    public long getNumPasses() {
	    	return this.numPasses;
	    }
	}
	
	class TestResult {
		public int numThreads;
		public double overallTimeTaken,
			threadTimeTaken,
			avgTimePerPass;
	}
	
	public TestResult testThreads(int numThreads) {
		TestThread [] threads = new TestThread[numThreads];
		
		for (int i = 0; i < numThreads; i++) {
			threads[i] = new TestThread(OfflineDocumentSpecs.getReliefwebHomeSpec());
		}

		long overallStartMillis = System.currentTimeMillis();            

		for (int i = 0; i < numThreads; i++) {
			threads[i].start();
		}
		
		boolean allFinished = false;

		while (!allFinished) {
			allFinished = true;
			for (int i = 0; i < numThreads; i++) {
				if (threads[i].isAlive()) {
					allFinished = false;
				}
			}
		}

		long overallEndMillis = System.currentTimeMillis();            
		long overallTimeTaken = overallEndMillis - overallStartMillis;
		
		long threadTimeTaken = 0;
		long totalNumPasses  = 0;
		
		for (int i = 0; i < numThreads; i++) {
			threadTimeTaken += threads[i].getTimeTotal();
			totalNumPasses  += threads[i].getNumPasses();
		}

		long avgTimePerPass = threadTimeTaken / totalNumPasses;
		
		/*
		System.out.println("Number of threads  : " +numThreads);
		System.out.println("Overall time taken : "+
				(double)overallTimeTaken/1000+" s");
		System.out.println("Thread  time taken : "+
				(double)threadTimeTaken/1000+" s");
		System.out.println("Mean time per pass : "+
				(double)avgTimePerPass/1000+" ms");
		System.out.println();
				*/
		
		TestResult result = new TestResult();
		result.numThreads = numThreads;
		result.overallTimeTaken = (double)overallTimeTaken / 1000;
		result.threadTimeTaken  = (double)threadTimeTaken  / 1000;
		result.avgTimePerPass   = (double)avgTimePerPass   / 1000;
		
		return result;
	}

	public void testConcurrency() {
		// warm up
		testThreads(10);

		Vector results = new Vector();
		
        LOG.debug(
				"Number of threads, " +
				"Overall time taken, " +
				"Thread  time taken, " +
				"Mean time per pass");

		double sum_x = 0.0, sum_y = 0.0;
		int n = 0;
		
		for (int numThreads = 1; numThreads <= 10; numThreads++) {
			TestResult result = testThreads(numThreads); 
			results.add(result);
            LOG.debug(
					result.numThreads + ", " +
					result.overallTimeTaken + ", " +
					result.threadTimeTaken  + ", " +
					result.avgTimePerPass);
			sum_x += result.numThreads;
			sum_y += result.avgTimePerPass;
			n++;
		}
		
		/* We 'plot' the number of threads (x) against average time per
		 * filter pass (y), and calculate a linear regression. The gradient
		 * of the average line, expressed as a fraction of the average time
		 * per pass, represents the marginal cost of another simultaneous
		 * thread. If this is zero, then extra threads are free, and we
		 * have infinite concurrency. If it's one, then extra threads cost
		 * as much as the first, and we have no benefit from concurrency.
		 * If greater than one, concurrency is worse than no concurrency.
		 * Our target is less than 0.5.    
		 */ 
		
		double mean_x = sum_x / n;
		double mean_y = sum_y / n;
		
		double gradient_sum = 0.0;
		int    gradient_n   = 0;
		
		for (int i = 0; i < results.size(); i++) {
			TestResult r = (TestResult)(results.get(i));
			/* if r.numThreads for this result equals the mean, then
			 * the gradient is undefined (division by zero) and this
			 * point must be ignored. 
			 */
			if (r.numThreads != mean_x) {
				double gradient = (r.avgTimePerPass - mean_y) / 
					(r.numThreads - mean_x);
				gradient_sum += gradient;
				gradient_n++;
			}
		}
		
		double gradient_avg = gradient_sum / gradient_n;
		double marginal_cost_factor = gradient_avg / mean_y; 
		
        LOG.debug("Marginal cost factor is "+marginal_cost_factor);
		
		assertTrue("Not enough concurrency: marginal cost factor "+
				"is greater than 0.5 ("+marginal_cost_factor+")",
				marginal_cost_factor < 0.5);
	}

    public static Test suite() {
        return new TestSuite(TestConcurrency.class);
    }
}
